;;;-*- Mode: Lisp; -*-
;;; Ystok-Widgets - LispWorks CAPI-based GUI framework
;;; Copyright (c) 2003-2010 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; CAPI internals-related utils, fixes, and (re)definitions

(in-package :ystok.widgets)

;;; win32:get-system-metrics argument constants; values are in pixels.
;;; LW5 CAUTION: In delivery, the structure class WIN32::SYSTEM-METRICS should be
;;;              mentioned in the :symbols-to-keep-structure-info list!
;;; Adopt other code for LW Linux version appropriately.
;#+win32 (defconstant SM_CXVSCROLL 2)	; width of a vertical scroll bar
;#+win32 (defconstant SM_CYHSCROLL 3)	; height of a horizontal scroll bar

(define-symbol-macro *hscroll-bar-height*
 ;;; (win32:get-system-metrics SM_CYHSCROLL)
  #+(and lispworks4 win32)
  (WIN32::SYSTEM-METRICS-CYHSCROLL CAPI-WIN32-LIB::*SYSTEM-METRICS*)
  #+(and (not lispworks4) mswindows)
  (slot-value CAPI-WIN32-LIB::*SYSTEM-METRICS* 'WIN32::CYHSCROLL)
  #-(or win32 mswindows)
  18)							; Linux/MacOS guess

(define-symbol-macro *vscroll-bar-width*
 ;;; (win32:get-system-metrics SM_CXVSCROLL)
  #+(and lispworks4 win32)
  (WIN32::SYSTEM-METRICS-CXVSCROLL CAPI-WIN32-LIB::*SYSTEM-METRICS*)
  #+(and (not lispworks4) mswindows)
  (slot-value CAPI-WIN32-LIB::*SYSTEM-METRICS* 'WIN32::CXVSCROLL)
  #-(or win32 mswindows)
  18)							; Linux/MacOS guess

;#+win32 (defconstant HWND_TOP        0)
;#+win32 (defconstant HWND_BOTTOM     1)
#+win32 (defconstant HWND_TOPMOST   -1)
#+win32 (defconstant HWND_NOTOPMOST -2)

#+win32 
(defmacro to-unsigned32 (long)
 ;;; Used for decoding the result of win32:get-window-long
  `(ldb (byte 32 0) ,long))

#+win32 
(defun to-signed32 (long)
 ;;; Used to specify the third argument of win32:set-window-long
  (if (logtest #.(ash 1 31) long)	; negative
      (dpb long (byte 32 0) -1)
      long))				; positive

(defconstant GWL_EXSTYLE -20)
(defconstant WS_EX_TOPMOST #x08)
#+(and win32 lispworks4.3) (defconstant WS_EX_TOOLWINDOW #x80)
;; Provide a constant of typesys:int32 instead of bignum #x80000000!
#+win32 (defconstant WS_POPUP #x80000000) ;-2147483648)
;;; Samples:
;;;	(win32:get-window-long handle win32:GWL_STYLE)
;;;	(win32::get-window-ulong handle win32:GWL_STYLE)
;;;	(win32:set-window-long handle win32:GWL_STYLE WS_POPUP) ; WS_TABSTOP
;;;	(win32:set-window-long handle GWL_EXSTYLE WS_EX_TOOLWINDOW)

;;; The double-click time is the maximum number of milliseconds that may occur
;;; between the first and second click of a double-click.
;;; In LW, internal-time-units-per-second=1000, i.e. internal-unit is millisecond.
#+win32 
(ffc:define-win32api |GetDoubleClickTime| () win32:UINT :module "user32")

(defvar *double-click-duration*			; in internal units
         #+win32 (* (|GetDoubleClickTime|) #.(/ internal-time-units-per-second 1000))
         #-win32 500)

(lw:define-action "When starting image" "Initialize system metrics"
    (lambda () 
      #+win32 (CAPI-WIN32-LIB::INIT-SYSTEM-METRICS)	; Is not called auto if %dl% = 5!
      (setq *double-click-duration*
            #+win32 (* (|GetDoubleClickTime|) #.(/ internal-time-units-per-second 1000))
            #-win32 500)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  UTILS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defgeneric ambient-switchable-layout-and-child (pane &optional child)
 ;;; Find a (grand)parent layout for the pane.
  ;; Values: 1> Instance of capi:switchable-layout or CAPI-INTERNALS:LIBRARY-TAB-LAYOUT
  ;;	     2> Switchable child, e.g. pane on which the object resides,
  ;;		or nil if impossible to determine.
  (:method (self &optional child) (declare (ignore child)) nil)
  (:method ((self capi:switchable-layout) &optional child) (values self child))
  (:method ((self CAPI::BASIC-ELEMENT) &optional child)		; recurse
   (declare (ignore child))
   (ambient-switchable-layout-and-child (capi:element-parent self) self))
)

(defun get-active-pane-interface (object)
 ;;; Helper called by active-pane-* and others.
  ;; Args: object Either focus pane supplied by a selection-callback
  ;;              or top-level interface supplied by enabled-function.
  ;; Values: 1) An active pane 2) Its top-level interface.
  (let* ((interface (if object
                        (capi:top-level-interface object)
                        (current-top-level-interface)))
         (pane (if (and object
                        (not (capi:top-level-interface-p object))
                        (capi:pane-has-focus-p object))
                   object
                   (capi::pane-descendant-child-with-focus interface))))
    (yl:debug-format :ywidgets "<< %get-active-pane-interface ~s ~s" pane interface)
    (values pane interface)))

;;; WORKAROUND:
;;;	Setting capi:choice-selection programmaticly do change visual text
;;;	but does not ensure the value (capi:text-input-pane-text choice) until you
;;;	place the cursor there!
(let ((lw:*handle-warn-on-redefinition* #+lispworks4 :quiet #-lispworks4 :warn))
  (defmethod (setf CAPI:CHOICE-SELECTION) :after (selection
                                                  (choice capi:text-input-choice))
    (setf (capi:text-input-pane-text choice)
          (if selection
              (funcall (capi:collection-print-function choice)
                       ;(capi:get-collection-item choice (capi:choice-selection choice)))
                       (capi:choice-selected-item choice))
              +null-string+))))

(defun find-data-in-collection (collection data &key (test #'eql))
 ;;; Values: 1> index in collection or nil
  ;;	     2>	item itself
  ;; NB: collection-test-function is a comparison function between two items, not data!
  ;;     key-function is applied to data, not to an item, during capi:search-for-item
  (let ((index 0))
    (declare (fixnum index))
    (capi:map-collection-items collection
     (lambda (item)
       (when (funcall test (capi::collection-item-data collection item) data)
         (return-from find-data-in-collection (values index item)))
           ;(values (capi:search-for-item collection item) item)))
       (incf index)))))

(defun find-pane (parent class &rest args &key test (leaf-only t) visible reverse
                                          &allow-other-keys)
 ;;; Args: test Applied only after the type is checked.
  ;;	   args Plist; if specified, all property value of plist must match
  ;;		the same property value of the child.
  ;; 	   leaf-only = T
  ;;		 layout' ( capi:tab-layout),   -
  ;;		capi::pane-with-implementation, e.g. text-input-range, 
  ;;		  - capi::pane-implementation-mixin,
  ;;		e.g. library-text-input-range.
  ;; NB: The find-pane-... family starts working only after capi:interface-display,
  ;;     but not within initialize-instance!
  (let ((plist (remove-properties args '(:test :leaf-only :visible :reverse))))
    (capi:map-pane-descendant-children parent
     (lambda (child)
       (and (typep child class)
            (or (null test) (funcall test child))
            (do ((rest plist (cddr rest)))
                ((null rest) t)
              (unless (eql (capi:capi-object-property child (first rest)) (second rest))
                (return nil)))
            (return-from find-pane child)))
     :reverse reverse :visible visible
     :leaf-only leaf-only)))

(defun find-pane-by-name (parent name &rest plist)
 ;;; Args: plist If specified, all property value of plist must match
  ;;		 the same property value of the child.
  (capi:map-pane-descendant-children parent
    (lambda (child)
      (and #-lispworks4.3 (capi:capi-object child)
           #+lispworks4.3 (typep child 'capi:capi-object)
           (or (null name) (eq (capi:capi-object-name child) name))
           (do ((rest plist (cddr rest)))
               ((null rest) t)
             (unless (eql (capi:capi-object-property child (first rest)) (second rest))
               (return nil)))
           (return-from find-pane-by-name child)))))

#+win32
(ffc:define-win32api |FlashWindow| ((handle win32:HWND) (reverse win32:BOOL))
                     win32:BOOL
                     :module "user32")
(defun flash-window (interface &optional (reverse t))
 ;;; Flash taskbar indicator and the window border as active
  ;; Args: reverse True = inverse current flash mode
  ;;		   Fasle = switch off flashing
  ;; Value: Old flash mode
  ;; NB:   ,    ?
  (when-let (handle (object-handle interface))
    #+win32 (|FlashWindow| handle (if reverse 1 0))))

(defgeneric implemented-pane (pane-or-implementation)
 (:method (pane) pane)
 (:method ((implementation capi::pane-implementation-mixin))
  (CAPI::PANE-IMPLEMENTED-PANE implementation)))		; go back to pane itself

(defun interface-owner (interface)
 ;;; Arg: interface  Modeless or modal interface opened with capi:display/display-dialog.
  ;; Value: NIL or a (top-level) interface, the same as passed by the :owner initarg;
  ;;        possibly a instance child-interface.
  (CAPI::INTERFACE-OWNER interface))

(defgeneric message-area-format (pane format-control &rest format-arguments)
 ;;; NB: Method #<(setf capi:titled-object-message) (t interface)>
  ;;     seems to replace newline characters by spaces in the resulting string itself.
 (:method ((pane capi:message-pane) string &rest format-arguments)
  (setf (capi:title-pane-text pane)
        #1=(if format-arguments (apply #'format nil string format-arguments) string)))
 (:method ((pane capi:element) string &rest format-arguments)
  (let ((interface (capi:top-level-interface pane)))
    (when (capi:interface-message-area interface)
      (setf (capi:titled-object-message interface) #1#))))
 (:method (pane string &rest format-arguments)
   (when-let (interface (current-top-level-interface))
     (when (capi:interface-message-area interface)
       (setf (capi:titled-object-message interface) #1#)))) )

(defgeneric object-handle (object)
 (:method (object) 0)
 (:method ((object integer)) object) 
 (:method ((object capi:simple-pane)) (capi:simple-pane-handle object))

 (:method ((object capi:text-input-range))
  ;; Representation and hanlde are only available through implementation
  (if-bind (implementation (CAPI::PANE-IMPLEMENTATION object))
    (capi:simple-pane-handle implementation)
    0))

 #+win32 (:method ((object win32:r-window))
          (win32::r-window-hwnd object))
 #+win32 (:method ((object (eql t)))
          ;; OT1H, (win32:get-active-window) does not work here - strange!
          ;; OTOH (ww::find-topmost-window nil nil) may not be equal to actually
          ;; active window, but independent topmost instead
          (object-handle ww::*active-window*)) )

(defun object-preference-path (object)
 ;;; CAUTION: Collisions are possible due to embedded pane with similiar names!
  ;; CAUTION: capi-object-name may be NIL if the object was not created,
  ;;          so its type symbol could potentially uninterned is appended!
  (let ((face (capi:top-level-interface object)))
    (macrolet ((preference-environment-path (key &optional subkey)
                 (with-gensyms (symbol package pane-name)
                   `(let* ((,symbol ,key)
                           (,package (symbol-package ,symbol))
                           (,pane-name ,subkey))
                      (list* "Environment"
                             (if ,package (package-name ,package) "uninterned")
                             (symbol-name ,symbol)
                             (when ,pane-name (list (string ,pane-name))))))))
      (preference-environment-path
       (if (capi:top-level-interface-save-geometry-p face)
           (capi:top-level-interface-geometry-key face)
           (type-of face))
       (unless (eq object face)
         (or (capi:capi-object-name object)	; usually coincides with slot name
             (type-of object)))))))

(defgeneric object-representation (object)
 (:method (object)
  (CAPI::REPRESENTATION object))
 (:method ((object CAPI::PANE-WITH-IMPLEMENTATION))
  ;; Representation is only available through implementation
  (when-let (implementation (CAPI::PANE-IMPLEMENTATION object))
    (CAPI::REPRESENTATION implementation))))

;(defun preference-environment-path (symbol &optional pane-name)
;  (list* "Environment" (package-name (symbol-package symbol)) (symbol-name symbol)
;         pane-name))

(defun raise-window (interface &optional activate)
 ;;; Allows control whether put the interface into focus
  #+win32 (win32:set-window-pos (capi:simple-pane-handle interface)
            0 0 0 0 0						 ; 0 = HWND_TOP
            (logior #.(logior win32:SWP_NOSIZE win32:SWP_NOMOVE) ; #x04 = SWP_NOZORDER
            (if activate 0 win32:SWP_NOACTIVATE)))
  #-win32 (declare (ignore activate))
  #-win32 (capi:raise-interface interface))

(defun switch-activate-pane (pane)
 ;;; Switch to the correct child if the pane is within a switchable-layout then activate.
  ;; Q: Do we need to invoke a selection-callback if the ambient switchable-layout is 
  ;;	actually capi-internals:library-tab-layout?
  (multiple-value-bind (layout child)
      (ambient-switchable-layout-and-child *erroneous-pane*)
    (when (and child
               (neq (capi:switchable-layout-visible-child layout) child)
               (member child (capi:switchable-layout-switchable-children layout)))
      (setf (capi:switchable-layout-visible-child layout) child)))
      #|(when (capi:callbacks-selection-callback tab-layout)
          (CAPI::CAPI-COLLECTION-CALLBACK tab-layout position
                 :index position
                 :callback-accessor 'capi:callbacks-selection-callback))|#
  ;(show-raise-restore (capi:top-level-interface pane))
  (capi:activate-pane pane))

#+unused
(defun undo-buffer-empty-p (buffer)
 ;;; Test whether the editor "Undo" command will do nothing
  ;; Use the accessor EDITOR::BUFFER-UNDO to get the undo slot of the editor buffer.
  ;; The value is usually an object of type SYSTEM:RING (but can be NIL initially).
  ;; Args:  buffer - editor's buffer.
  ;; Value: True iff the "Undo" command will do nothing
  (not (and (editor:buffer-modified buffer) (editor:buffer-writable buffer))))
  #|(let ((undo (EDITOR::BUFFER-UNDO buffer)))
    (if (sys::ringp undo)
        ;; STUB: The below internals are absent in LW6. Sometimes not working on LW4.4
        #-lispworks6
        (let ((bound (sys::ring-bound undo)))
          (or (minusp bound) (= (sys::ring-v-pop undo) bound)))
        #+lispworks6
        (<= (sys::ring-length undo) 0)			; actually does not work!
        t)))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  FOCUS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod capi:accepts-focus-p ((self CAPI::CAPI-GENERIC-LAYOUT-DIVIDER))
 ;;; By default, capi:accepts-focus-p is true for CAPI::CAPI-GENERIC-LAYOUT-DIVIDER
  ;; - too bad!
  nil)

(defgeneric find-pane-to-focus (parent &optional test)
 ;;; Search for a pane within the parent that can accept focus via capi:set-pane-focus
  ;; Args: test Predicate that the resulting pane must satisfy, e.g. enabled
 (:method (parent &optional test)
  (declare (ignore parent test))
  nil)
 (:method ((parent capi:simple-pane) &optional test)
  (when (and (capi:accepts-focus-p parent) (imp test (funcall test parent)))
    parent))
 (:method ((parent CAPI::PANE-WITH-CHILDREN) &optional test)
  (when-let (pane (CAPI::PANE-DEFAULT-FOCUS-DESCENDANT-CHILD parent))
    (when (imp test (funcall test pane))
      pane)))
  #|(find-pane parent 'CAPI::SIMPLE-ELEMENT :leaf-only t
             :visible t 	; when nil, cannot find editing pane within pinboard!
             :reverse t
             :test (if test
                       (lambda (pane) (and (capi:accepts-focus-p pane)
                                           (funcall test pane)))
                       #'capi:accepts-focus-p)))|#
 (:method ((parent capi:switchable-layout) &optional test)
  (find-pane-to-focus (capi:switchable-layout-visible-child parent) test)) )

 #|(:method (interface (parent capi:extended-selection-tree-view) &optional test)
  ;; Within an this tree, only an instance of CAPI::GENERIC-TREE-VIEW accepts focus.
  ;; But it excluded from the ordinary search due to :leaf-only t.
  ;(find-pane-to-focus (CAPI::PANE-IMPLEMENTATION parent) test)) 
  ;(when (capi:accepts-focus-p parent)
  (when-let (pane (CAPI::PANE-DEFAULT-FOCUS-DESCENDANT-CHILD parent))
    (when (imp test (funcall test parent))
      pane))))|#

(defun show-raise-restore (interface)
 ;;; Bring hidden or iconic interface to front restoring its to last normal or maximized
  ;; CAPI Bugs:
  ;;  1) Maximized and hidden window becomes visible but restored to normal size.
  ;;  2) The window is not always brought to front - we use capi:activate-pane for this.
  ;; Doc states: ...depending on how the interface was visible in the past.
  (case (capi:top-level-interface-display-state interface)
    ((:hidden :iconic)
     (setf (capi:top-level-interface-display-state interface) :restore))
    (otherwise
     (capi:raise-interface interface)))
  (when-let (pane (find-pane-to-focus interface))
    (capi:activate-pane pane)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  MULTIPROCESSING  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun wait-execute-with-interface (interface function &rest args) ; &key timeout)
 ;;; Determine if the function called using execute-with-interface has completed.
  ;;  (Based on ideas provided by W.Humeniuk and N.Levine)
  ;;  We sleep the parent thread making sure unwind-protect on the child operation,
  ;;  otherwise if it dies, we'll have locked the parent.
  ;; Value: Single value returned by function, or :abort, or :timeout.
  (let ((process mp:*current-process*)
        (timeout (if (eq function :timeout)
                     (prog1 (first args)
                       (setq function (second args)
                             args (cddr args))))))
    (capi:execute-with-interface interface
     (lambda (args)
       (let ((result :abort))
         (unwind-protect (setq result (apply function args))
           (when process			; test whether timeout has not expired
             (mp:process-interrupt process	; invoke return within the origial process
                 (lambda () (return-from wait-execute-with-interface result)))))))
     args)
    (if timeout
        (progn
          (mp:process-wait-with-timeout #1="Wait for EXECUTE-WITH-INTERFACE completion"
                                        timeout)
          (setq process nil)
          :timeout)
        (mp:process-wait #1# #'lw:false))))

;;; Auxiliary shared process, which loops over requests in its mailbox and evaluates them.
;;; Used for working around various bugs by means of process switching:
;;; 1) Tree-view click or double-click creating and displaying another interface.
;;; 2) Text-input-pane as editing pane - setting caret position just on creation.

(defvar %aux-mailbox% nil)				

(defun aux-funcall (function &rest args)
  (when (if %aux-mailbox%					; ensure-mailbox
            (not (and (mp:process-alive-p
                       (mp:mailbox-reader-process (orf %aux-mailbox% (mp:make-mailbox))))
                      (equal (mp:process-name (mp:mailbox-reader-process %aux-mailbox%))
                             #1="YWidgets auxiliary process")))
            (setq %aux-mailbox% (mp:make-mailbox)))
    (setf (mp:mailbox-reader-process %aux-mailbox%)
          (mp:process-run-function #1# (list :mailbox %aux-mailbox%)
           (lambda ()						; runner
             (loop (let ((form (mp:mailbox-read %aux-mailbox%)))
                     (yl:debug-format :ywidgets "~s -> s" %aux-mailbox% form)
                     (ignore-errors (apply (first form) (rest form)))))))))
  (mp:mailbox-send %aux-mailbox% (cons function args)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  FONT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun string-hint-length (arg)
 ;;; Adjusted length in characters used to define hints (:character, :text-width etc.)
  ;; relied on gp:get-font-average-width.
  ;; Args: arg String or fixnum.
  ;; Rationale: Characters are wider in some languages.
  (declare (optimize (speed 3) (safety 0) (hcl:fixnum-safety 0)))
  (let ((length (if (numberp arg) arg (length arg))))
    (declare (fixnum lenght))
    #-Russian length
    #+Russian (the fixnum (+ length
                             (the fixnum (ceiling (the fixnum (* length 22)) 100))))))

(let (alist)
 (defun bold-font (&optional (delta 0))
 ;;; Args: delta Increment or decrement to standard font size (in points).
  (orf (cdr-assoc delta alist)
       #+win32 (gp:find-best-font (first CAPI-WIN32-LIB::*SCREENS*)
                 (let ((description (gp:font-description
                                     CAPI-WIN32-LIB::*CONTROL-DEFAULT-FONT*)))
                   (apply #'gp:augment-font-description description
                          ;:family "MS Sans Serif"		; "MS Shell Dlg"
                          :weight :bold
                          (unless (zerop delta)
                            (list :size (+ delta (or (gp:font-description-attribute-value 
                                                      description :size) 10)))))))
       #-win32 (gp:make-font-description :family "Helvetica"
                                         :weight :bold
                                         :size (+ delta 10)))))
#+old
(let (font)
 (defun ensure-bold-font ()
   (orf font
        (gp:find-best-font (first CAPI-WIN32-LIB::*SCREENS*)
         #+win32 (gp:augment-font-description
                    ;(if pane (gp:font-description (capi:simple-pane-font pane))
                  (gp:font-description CAPI-WIN32-LIB::*CONTROL-DEFAULT-FONT*)
                  :family "MS Sans Serif"		; "MS Shell Dlg" looks ugly
                  :weight :bold)
         #-win32 #.(gp:make-font-description
                    :weight :bold :family "Helvetica" :size 10)))))

(defun merge-fonts (plist font)
 ;;; Args: plist Attributes (TODO: allow fdesc as well)
  ;;       font	 Font description, font, or graphics port
  ;; Value: Sturcture of type gp:font-description.
  ;; TODO:   plist  :size :s1/:s2/:e1/:e2,    
  ;;	   (+ (gp:font-description-attribute-value fd :size) +1/+2/-1/-2)
  (apply #'gp:augment-font-description
         (cond ((gp::font-description-p font) font)
               ((GP::FONT-P font) (gp:font-description font))
               (t (gp:font-description (gp:graphics-state-font (gp:get-graphics-state
                                                                font)))))
         plist))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  BUTTON  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Needed for :text-width hint to work

(let ((lw:*handle-warn-on-redefinition* #+lispworks4 :quiet #-lispworks4 :warn))
 (defmethod CAPI::MAXIMUM-STRING-LENGTH ((self capi:push-button))
   (CAPI::ELEMENT-TEXT-WIDTH self (capi:item-text self))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  MENU  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod (setf object-title) (value (self capi:menu) &optional parent)
  (declare (ignore parent))
  (setf (slot-value self 'CAPI::TITLE) value))

(defgeneric insert-menu (item capi-object &key after before)
 ;;; Args: after, before  Either are both menu objects or names (symbols)
  ;; Value: capi-object
 (:method ((item null) object &key after before)
  (declare (ignore after before))			; as item function return NIL,
  object)						; maybe the item is not featured
 (:method (item (interface capi:interface) &key after before)
  (setf (capi:interface-menu-bar-items interface)
        (yl:insert (if (or (symbolp item) (functionp item))
                       (funcall item interface)
                       item)
                   (copy-list (capi:interface-menu-bar-items interface))
                   :key (if (symbolp (or after before))
                            #'capi:capi-object-name #'identity)
                   :after after  :before before))
  interface)
 (:method (item (menu capi:menu) &key after before)
  (setf (capi:menu-items menu)
        (yl:insert (if (or (symbolp item) (functionp item))
                       (funcall item (capi:top-level-interface menu))
                       item)
                   (copy-list (capi:menu-items menu))
                   :key (if (symbolp (or after before))
                            #'capi:capi-object-name #'identity)
                   :after after  :before before))
  menu)
 (:method ((item capi:menu-object) (component capi:menu-component) &key after before)
  (if-bind (position (or after before))
    (setf (capi:collection-items component)
          (yl:insert item
                     (coerce (capi:collection-items component) 'list)
                     :key (if (symbolp position) #'capi:capi-object-name #'identity)
                     :after after  :before before))
    (capi:append-items component (list item)))			; simply push to the end
  component) )

(defgeneric pane-post-menu (pane x y &optional gesture)
 ;;; Work-horse of popup-menu and also used in output-pane input-model
  ;; Args: gesture Gesture-spec or :keyboard-post-menu if invoked by means of
  ;;               the accelerator (SHIFT+F10 on Windows),
  ;;		   or data coming from the popup-menu function - the selected
  ;;		   object/item if there is a selection,
  ;;		   or NIL.
  ;; Usage:
  ;;  '((:post-menu          pane-post-menu)
  ;;    (:control-post-menu  pane-post-menu :control)
  ;;    (:keyboard-post-menu pane-post-menu))
 (:method (pane x y &optional gesture)
  (declare (ignore pane x y gesture))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  TOOLBAR  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod (setf object-title) :before (value (self capi:toolbar) &optional parent)
 ;;; CAUTION: This method should be invoked before displaying, e.g. in initialize-instance!
  (when (typep (orf parent (implemented-pane (capi:element-parent self)))
               'capi:docking-layout)
    (when-let* ((items (capi:docking-layout-items parent))
                (tuple (or (assoq self items)		; (file-toolbar :title #L"File")
                           (assoq (capi:capi-object-name self) items))))
      (setf (getf (rest tuple) :title) value))))
    ;(call-next-method)))

(let ((lw:*handle-warn-on-redefinition* #+lispworks4 :quiet #-lispworks4 :warn))
 (defmethod CAPI:SET-BUTTON-PANEL-ENABLED-ITEMS
            ((collection CAPI::TOOLBAR-BUTTON-GROUP)
             &key enable disable key
                  (set nil supplied-p)
                  (test (or (capi:collection-test-function collection) #'eql)))
  (declare (ignore key))
  (map nil (lambda (data button)
             (when (CAPI::TOOLBAR-OBJECT-P data)
               (setq data (capi:item-data button)))
             (cond (set 
                    (setf (capi:simple-pane-enabled button)
                          (not (member data disable :test test))))
                   (supplied-p
                    (setf (capi:simple-pane-enabled button)
                          (member data enable :test test)))
                   ((member data enable :test test)
                    (setf (capi:simple-pane-enabled button) t))
                   ((member data disable :test test)
                    (setf (capi:simple-pane-enabled button) nil))))
                   ; otherwise: do not change the enabled status
       (capi:collection-items collection)	; => vector of either buttons or data
       (CAPI::TOOLBAR-BUTTON-GROUP-BUTTONS collection))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  CURSOR  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declaim (inline cursor-eql set-cursor restore-cursor))

#+(and lispworks4 win32)
(ffc:define-win32api |LoadCursorFromFile| ((lpFileName :pointer)) win32:DWORD
                     :encoding :dbcs :module "user32")

(defun load-cursor-from-file (pathname)
 ;;; Value: Windows handle (or 0 in the file is not found) on LW4
  ;;        or a cursor object on LW5.
  #+(and lispworks4 win32)
  (fli:with-foreign-string (ptr element-count byte-count
                                :external-format ffc:*foreign-external-format*)
      (namestring pathname)
    (declare (ignore element-count byte-count))
    (|LoadCursorFromFile| ptr))
  #-(and lispworks4 win32)
  (capi:load-cursor (namestring pathname)))

(defun cursor-eql (pane cursor)
  #+(and lispworks4 win32) (eql (capi-win32-lib::r-output-pane-cursor
                                 (capi-internals:representation pane)) cursor)
  #-(and lispworks4 win32) (eq (capi:simple-pane-cursor pane) cursor))

(defun set-cursor (pane value)
  (setf #+(and lispworks4 win32) (capi-win32-lib::r-output-pane-cursor
                                  (capi-internals:representation pane))
        #-(and lispworks4 win32) (capi:simple-pane-cursor pane)
        value))

(defun restore-cursor (pane)
 ;;; Restore standard cursor
  #+(and win32 lispworks4)
  (let ((r (capi-internals:representation pane)))
    (when (capi-win32-lib::r-output-pane-cursor r)
      (setf (capi-win32-lib::r-output-pane-cursor r) nil)))
  #-(and win32 lispworks4)
  (when (capi:simple-pane-cursor pane)
    (setf (capi:simple-pane-cursor pane) nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  LAYOUT  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Pane instances that can be incorporated into layout description for
;;; auto-expandable space

(defun h-space (&optional (min +h-gap+) max)
  (make-instance 'CAPI::DUMM-PANE :visible-min-width min :visible-max-width max))

(defun v-space (&optional (min +v-gap+) max)
  (make-instance 'CAPI::DUMM-PANE :visible-min-height min :visible-max-height max))

;;;;;;;;;;;;;;;;;;;;;;;;;  CAPI::OPTION-PANE-CHOICE-MIXIN  ;;;;;;;;;;;;;;;;;;;;;;;

(defun set-dropped-width (panes &optional (width nil supplied-p))
 ;;; Args: panes  Instance list or instance of CAPI::OPTION-PANE-CHOICE-MIXIN
  ;;	   width  Pixels, :text-width, or any allowed in hint.
  ;; Value: CB_ERR (-1) if failed
  ;; Usage: Should be called by :before/:after method on capi:interface-display
  ;;        or on creating tab/swith-layout page and after changing items.
  ;; NB: Does not reposition the poped up list to the left if clipped outside the screen.
  ;; CAUTION: When there is a separator-item,
  ;;	      - print-function must be applicable to it (for hints like :text-width),
  ;;          - enabled-positions (if any) must count it!
  #+win32
  (dolist (pane (listify panes))
    (when (if supplied-p width (setq width (getf (pane-args pane) :min-width)))
      (win32:send-message
       (object-handle pane)				; HWND
       #x0160						; MSG = CB_SETDROPPEDWIDTH
       (max (capi:with-geometry pane			; WPARAM
              (or capi:%width% capi:%min-width% 0))
            (+ (string-hint-length (CAPI::INTERPRET-HINT-VALUE
                                    pane :visible-min-width width))
               (let ((count (capi:option-pane-visible-items-count pane)))
                 (if (and (iplusp count)			; can be :default
                          (<= (capi:count-collection-items pane) count))
                     0
                     *vscroll-bar-width*))))		;(ash *vscroll-bar-width* 1)
       0))))						; LPARAM

;(defmethod CAPI::CREATE-GEOMETRY-CACHE :after ((pane CAPI::OPTION-PANE-CHOICE-MIXIN))
;  (unless (eq (capi:collection-items pane) +items-uninitialized+) ; prevent fault setting
;    (set-dropped-width pane)))

(defun show-dropdown (pane &optional (state t))
 ;;; Args: pane  Instance of CAPI::OPTION-PANE-CHOICE-MIXIN
  ;;	   state True = show the list box, False = hide it.
  #+win32
  (win32:send-message (object-handle pane)		; HWND
                      win32:CB_SHOWDROPDOWN		; MSG
                      (if state 1 0)			; WPARAM ::= TRUE or FALSE
                      0))				; LPARAM - not used

;; Workaround for displaying images in drop-down-list of capi:option-pane

#+(and win32 lispworks4)
(defmethod GP::IMAGE-BACKGROUND-COLOR ((object-representation CAPI-WIN32-LIB::R-COMBOBOX))
  'win32:color_window)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  GRAPH  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; WORKAROUND: Patch needed for graphs with edges-selectable-p = t
;;;  The native function works for all directions except up and down where it signals.
;;;  We are ignoring pressing UP or DOWN key.
(let ((lw:*handle-warn-on-redefinition* :quiet))
  (defmethod CAPI:GRAPH-NODE-IN-EDGES ((edge capi:graph-edge))
    ())
)
#|;;; More subtle approach
(setf (fdefinition 'native-CAPI-GRAPH-NODE-ADJACENT-SIBLING)
      (fdefinition 'CAPI::GRAPH-NODE-ADJACENT-SIBLING))
(defun CAPI::GRAPH-NODE-ADJACENT-SIBLING (edge direction)
  ;; Args: direction -1=up, 1=down
  (if (and (typep edge 'capi:graph-edge) (eql (abs direction) 1))
      (graph-edge-adjacent-sibling (capi:graph-object-object (capi:graph-edge-from edge))
                                   (capi:graph-object-object (capi:graph-edge-to edge))
                                   (if (plusp direction) :down :up))
      (native-CAPI-GRAPH-NODE-ADJACENT-SIBLING edge direction)))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  RECTANGLE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Concept implemented slightly different from GP's
;;; Aka GEOMETRY, which is compatible with gp:draw-rectangles and gp:with-graphics-mask.

(eval-when (:compile-toplevel :load-toplevel :execute)
  (setf (fdefinition 'make-rect) (fdefinition 'list)))
;(defmacro make-rect (left top width height) `(list ,left ,top ,width ,height))

;;; Is it dangerous to declare: fixnum top?
(defmacro rect-left (rect) `(the fixnum (first ,rect)))
(defmacro rect-top (rect) `(second ,rect))
(defmacro rect-width (rect) `(the fixnum (third ,rect)))
(defmacro rect-height (rect) `(the fixnum (fourth ,rect)))

(defmacro rect-bind ((left top &optional width height) rect &body body)
  (with-gensyms (r)
    `(let* ((,r ,rect)
            ,@(when left `((,left (rect-left ,r))))
            ,@(when top `((,top (rect-top ,r))))
            ,@(when width `((,width (rect-width ,r))))
            ,@(when height `((,height (rect-height ,r)))))
       (declare (fixnum ,@(when left `(,left)) ,@(when top `(,top))
                        ,@(when width `(,width)) ,@(when height `(,height))))
       ,@body)))

;;; Working with simple translation while drawing on screen

(defun translate-point (port x y)
 ;;; A la gp:transform-point when the port transform is surely a simple translation
  (let ((translation (cddddr (gp:graphics-port-transform port))))
    (values (+ x (first translation))
            (+ y (second translation)))))

(defmacro translate-mask (port rect &key left top width height)
 ;;; Convert the rect area or the coordinates given according to port translation
 ;;; Args: left top width height  Specify alternative bounds in internal coordinates
  (with-gensyms (translation)
    `(let ((,translation (cddddr (gp:graphics-port-transform ,port))))
       (,(if height 'list 'list*)
        (+ ,(or left `(rect-left ,rect)) (first ,translation))
        (+ ,(or top `(rect-top ,rect)) (second ,translation))
        ,@(cond (height `(,(or width `(rect-width ,rect))
                          ,height))			; do not reuse rect structure
                (width `(,width (cdddr ,rect)))		; reuse last cons
                (t `((cddr ,rect))))))))		; reause last two conses

(defmacro with-translated-mask ((port rect &key left top width height) &body body)
 ;;; Hacked version: don't save old mask, mask-x and mask-y
  ;; Args: rect  Must a variable or nil
  ;;       left top width height  Specify alternative bounds in internal coordinates
  (with-gensyms (state translation)
    `(let* ((,state (gp::graphics-port-graphics-state ,port))
            (,translation (cddddr (gp:graphics-state-transform ,state))))
       (unwind-protect
           (progn
             (setf (gp:graphics-state-mask ,state)
                   (,(if height 'list 'list*)
                    (+ ,(or left `(rect-left ,rect)) (first ,translation))
                    (+ ,(or top `(rect-top ,rect)) (second ,translation))
                    ,@(cond (height `(,(or width `(rect-width ,rect))
                                      ,height))		    ; do not reuse rect structure
                            (width `(,width (cdddr ,rect))) ; reuse last cons
                            (t `((cddr ,rect))))))	    ; reuse last two conses
             ,@body)
         (setf (gp:graphics-state-mask ,state) nil)))))	  ; old mask is always nil

;;; Working with full-fledged transform, e.g. for printing

(defmacro transform-mask (port rect &key left top width height)
 ;;; Args: left top width height  Specify alternative bounds in internal coordinates
  (if (or left top width height)
      `(multiple-value-list (gp:transform-area (gp:graphics-port-transform ,port)
                                               ,(or left `(rect-left ,rect))
                                               ,(or top `(rect-top ,rect))
                                               ,(or width `(rect-width ,rect))
                                               ,(or height `(rect-height ,rect))))
      `(multiple-value-list (apply #'gp:transform-area
                                   (gp:graphics-port-transform ,port)
                                   ,rect))))

(defmacro with-transformed-mask ((port rect &rest args) &body body)
 ;;; Hacked version: don't save old mask, mask-x and mask-y
  ;; Args: rect  Must a variable or nil
  ;;       left top width height  Specify alternative bounds in internal coordinates
  (with-gensyms (state)
    `(let ((,state (gp::graphics-port-graphics-state ,port)))
       (unwind-protect
           (progn
             (setf (gp:graphics-state-mask ,state) (transform-mask ,port ,rect ,@args))
             ,@body)
         (setf (gp:graphics-state-mask ,state) nil)))))	  ; old mask is always nil

(pushnew :ywidgets *features*)

;#+deprecated
;(setf (fdefinition 'find-pane-with-focus)
;      (fdefinition 'CAPI::PANE-DESCENDANT-CHILD-WITH-FOCUS))
#|(defun find-pane-with-focus (parent)
  ;; Q: Does not work for text-input-choice or objects within a pinboard?
  (capi:map-pane-descendant-children parent
    (lambda (child) (when (capi:pane-has-focus-p child)
                      (return-from find-pane-with-focus child)))
    :visible t :leaf-only nil))|#
