;;;-*- Mode: Lisp; -*-
;;; Ystok-Widgets - LispWorks CAPI-based GUI framework
;;; Copyright (c) 2003-2010 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Confirmer interface and various promters

(in-package :ystok.widgets)

;; Passed as :popup-args argument to CAPI standard prompts, e.g.
;; (prompt-with-list ... :interaction :multiple-selection
;;                   :popup-args (list* :owner interface :window-styles '(:resizable)
;;                                      +prompt-button-titles+))
(defconstant +prompt-button-titles+
  #+English '(:no-button nil)				; is OK everywhere OK?
  #-English `(:all-button ,+all-string+			; "All"
              :no-button nil				; or +none-string+ = "None"
              :cancel-button ,#L"Cancel"))

(defparameter *confirmer-buttons-aplist*
 `((:ok		:text #L"OK"   :mnemonic nil)		; don't auto underline
   (:cancel	:text #L"Cancel" :mnemonic nil)		; don't auto underline
   (:yes	:text #L"Yes")
   (:no		:text #L"No")
   (:all	:mnemonic-text #L"Yes to &All")
   (:none	:mnemonic-text #L"N&o to All")
   (:help	:text #L"Help" :mnemonic #-Russian 0 #+Russian 4
		:callback-type :data-interface		; as for help-function
                :help-key #L"Open a detailed topic in the help viewer.")))

(defgeneric chooser-top-level-interface (object)
 (:method (chooser) nil)
 (:method ((chooser capi:simple-pane))
  (let ((face (capi:top-level-interface chooser)))
    (if (CAPI::PANE-CREATED-P face) face nil))) )		; not destroyed

;;; Functions to be called just before popup is displayed or destroyed/hidden.
;;; Args: owner Chooser or it interface
;;;       popup/confirmer/choice-board
;;;		Instance of popup, confirmer or pseudo-modal choice-interface,
;;;		which is going to receive or loose focus.
;;; NB: For a top-level owner class, specialized methods MUST BE defined.
;;; LW6 NB: It is unclear how to hanlde capi:non-focus-list-interface.

(defgeneric popup-before-activate (owner popup)
 ;;; Called just before displaying popup, confirmer, or pseudo-modal,
 ;;; in the owner process (or in the choice-board process when it is created first).
  ;; CAUTION: The owner's chooser slot may not be set yet.
 (:method (chooser popup)
  (when (and chooser (not (capi:top-level-interface-p chooser)))
    (when-let (owner (chooser-top-level-interface chooser))
      (popup-before-activate owner popup)))) )

(defgeneric modal-before-deactivate (owner confirmer &optional hide)
 ;;; Called just before destroying/hidding modal confirmer or pseudo-modal
  ;; in the process of the choice-board, of their common, or arbitrary.
  ;; Args: hide  True = choice-interface is about to be destroyed or made hidden.
  ;;		 False = choice-interface is about to be reinitialized for another choice.
  ;; CAUTION: Must not change the owner's chooser slot.
  ;; NB:    popup,      ,
  ;;	       popup.
 (:method (chooser confirmer &optional (hide t))
  (when (and chooser (not (capi:top-level-interface-p chooser)))
    (when-let (owner (or (chooser-top-level-interface confirmer)	; => NIL if dead
                         (chooser-top-level-interface chooser)))
      (modal-before-deactivate owner confirmer hide)))) )

(defun confirmer-create-callback (confirmer)
 ;;; Used as :create-callback for various prompters (passed to CAPI::CONFIRMER)
  (popup-before-activate (interface-owner confirmer) confirmer))

;; Should be appended to the :popup-args argument when calling many CAPI confirmers
(defconstant +popup-args+ '(:create-callback confirmer-create-callback))

(defun help-function (&optional data interface)
 ;;; Used as the :help-function argument when calling confirmers
  ;; CAPI NB: popup-confirmer's passes no arguments to its help-function.
  (declare (ignore data))
  (when (orf interface (CAPI::CURRENT-DIALOG))
    (help-callback interface nil :topic (capi:help-key interface))))

(defmethod help-callback ((interface CAPI::CONFIRMER) (pane capi:button)
                          (type (eql :tooltip)) &optional help-key)
 ;;; Suppress stupid tooltips for confirmer buttons
  (declare (ignore pane type))
  (if (and *show-tooltips* (neq help-key (capi:capi-object-name pane)))
      (help-key-tooltip help-key)
      nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  CONFIRMER  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; It seems too late to invoke modal-before-deactivate in destroy-callback.
;;; As confirm-destroy-function does not work for dialogs,
;;; shall we augment capi:abort-dialog and capi:exit-dialog for callbacks?

(defclass confirmer (capi:interface)
 ((ok-check :initarg :ok-check)
  (ok-button :initarg ok-button))				; internal slot
 ;((icon-pane :initarg :image :initform nil)))
 (:default-initargs
  :create-callback (lambda (self)
                     (popup-before-activate (interface-owner self) self)
                     (%ok-check-confirmer self))
  ;:destroy-callback (lambda (self) (modal-before-deactivate (interface-owner self) self))
))

(defmethod %ok-check-confirmer (confirmer)
 ;;; Helper also used as create-callback
  ;; CAUTION: Mustnot be called before the interface is created!
  (when-let* ((ok-check (slot-value confirmer 'ok-check))
              (ok-button (slot-value confirmer 'ok-button)))
    (setf (enabled ok-button) (funcall ok-check confirmer))))

(defmethod capi:redisplay-interface :after ((interface confirmer))
  (%ok-check-confirmer interface))

(defmethod help-callback ((interface confirmer) (pane confirmer) type &optional help-key)
 ;;; Prevent recursive call
  (declare (ignore interface pane type help-key)))

(defun confirm (pane message &rest args
                &key title message-args
                     icon
                     (owner t)
                     focus #|(find-if (lambda (pane)
                                              (and (typep pane 'CAPI::SIMPLE-ELEMENT)
                                                  ;(not (typep pane 'capi:button-panel))
                                                   (capi:accepts-focus-p pane)))
                                            (if (typep pane 'capi:layout)
                                                (capi:layout-description pane)
                                                (listify pane))))|#
                     value-function ok-check 
                     (buttons (if (or value-function ok-check) '(:ok :cancel) ()))
                     default-button
                     help-key
                     (help-callback (if help-key 'help-callback nil))
                     (help-function (if help-key 'help-function nil))
                     (window-styles '(:resizable))
                     (flash t)
                &allow-other-keys)
 ;;; Display modal dialog, an instance of the confirmer class
  ;; Args: pane	    Instance of capi:simple-pane (including layout), list of panes,
  ;;		    or NIL.
  ;;	   message  Text for title pane at the very top; if is NIL, no message
  ;;	   message-args
  ;;		    Plist of graphic arguments. Font value can be one of
  ;;		       NIL, font description/object, or :bold -> (bold-font).
  ;;		    TODO: Get default font from Windows display Appearence settings.
  ;;	   title    Window title.
  ;;	   icon	    One of NIL, :info, :query, :warn, :error.
  ;;		    If true, also beeps. (Shall we have a distinguished beep parameter?)
  ;;	   buttons  List of button keywords and/or descriptions of the form
  ;;			(<keyword> . <plist>),
  ;;		    e.g. (:all :mnemonic-text "&Yes for all" :help-key "Proceed silently")
  ;;		    Must not menthion :help.
  ;;		    The button row is displayed at the very bottom.
  ;;	   default-button
  ;;		    Keyword designating a button or NIL meaning the first one is default.
  ;;	   owner    If eql to T (default), means current active window.
  ;;	   focus    If NIL, we do not pass :focus argument to capi:display-dialog
  ;;		    as it can determine focus fine itself.
  ;;	   value-function
  ;;		    Invoked only for OK button. It is passed the confirmer interface
  ;;		    (or another object specified in :callback-type for the :ok button).
  ;;		    If not supplied, the button name is retruned from confirmer.
  ;;		    If the value-function wants to indicate that the dialog cannot
  ;;		    return a value currently, then it should return a second non-NIL value.
  ;;	   ok-check Works only if OK button is present. It is invoked over the confirmer.
  ;;		    (The value-function is not called for it - this differs from
  ;;		    capi:popup-confirmer.)
  ;;	   help-key Help context designotor without tooltip, i.e. flat string
  ;;		    is treated as ready context.
  ;;		    If true, equips the dialog with the '?' and the Help button.
  ;;                Used to compute the help-key passed to help-callback;
  ;;       help-callback
  ;;		    Help callback for the entire confirmer interface.
  ;;       help-function
  ;;		    Callback for the Help button if any; signature: (data interface)
  ;; Values: 1) Keyword designating the button pressed, or NIL if canceled,
  ;;            or any value specified in (capi:exit-dilog value) call within callbacks.
  ;;         2) Successp (as returned by capi:display-dialog and capi:popup-confermer).
  ;; NB: The confirmer itself neither establishes an error handle over the value-function
  ;;	 call, nor over the ok-check call.
  ;; NB: Exit functions, e.g. ok-function, are missing.
  ;;	 By default the confirmer is exited by means of
  ;;     - for OK button, (compose capi:exit-dialog value-function),
  ;;     - for Cancel button, capi:abort-dialog,
  ;;     - for other buttons, capi:exit-dialog and returns the button name by default.
  ;;     More specific callbacks can be passed within the buttons argument.
  (flet ((%default (description arg &optional default &aux value)
           ;; Args: default - the last resort, i.e. the value supplied by aplist prevails.
           (cond ((and (consp description)
                       (neq (setq value (getf (cdr description) arg *unbound-value*))
                            *unbound-value*))
                  value)					; explicit in descriptions
                 ((neq (setq value (getf (cdr-assoq (first-or-self description)
                                                    *confirmer-buttons-aplist*)
                                         arg *unbound-value*))
                       *unbound-value*)
                  value)					; explicit in aplist
                 (t default))))
    (let* ((message (when message
                      (apply #'make-instance 'capi:title-pane
                             :name 'CAPI::CONFIRMER-MESSAGE	; for ENSURE-PANE-STRING
                             ;:visible-max-height nil
                             :text message
                             (if (eq (getf message-args :font) :bold)
                                 (list* :font (bold-font)
                                        (remove-properties message-args '(:font)))
                                 message-args))))
           (default-button (or default-button (first-or-self (first buttons))))
           (ok-button nil)
           (buttons
            (mapcar
             (lambda (description)
               (let* ((data (first-or-self description))
                      (default-p (eq data default-button))
                      ;(plist (if (consp description) (rest description) nil))
                      (text (%default description :text))
                      (button (apply #'make-instance 'capi:push-button
                                :data data
                                :name data
                                :default-p default-p
                                :cancel-p (eq data :cancel)
                                :visible-min-width '(max (:character 10) :text-width)
						;= `(:character ,(max 10 (length text)))
                                :callback (%default description
                                           :callback
                                            (case data
                                              (:cancel
                                               'capi:abort-dialog)
                                              (:ok
                                               (if value-function
                                                   (lambda (arg)
                                                     (multiple-value-bind (value condition)
                                                         (funcall value-function arg)
                                                       (unless condition
                                                         (capi:exit-dialog value))))
                                                   'capi:exit-dialog))
                                              (otherwise
                                               'capi:exit-dialog)))
                                :callback-type (%default description
                                                :callback-type (if (and (eq data :ok)
                                                                        value-function)
                                                                   :interface
                                                                   :data))
                                :text text
                                :mnemonic (%default description :mnemonic (when text 0))
                                :mnemonic-text (if text
                                                   nil
                                                   (%default description :mnemonic-text))
                                :help-key (%default description
                                                    :help-key (when help-key :default))
                                (when (consp description)
                                  (remove-properties (rest description)
                                   '(:text :mnemonic :callback :callback-type
                                     :help-key))))))
                 (when (eq data :ok)
                   (setq ok-button button))
                 ;(when (and default-p (not focus))
                 ;  (setq focus button))
                 button))
             (if (and help-key *show-help-button*)
                 (push-end `(:help :callback ,help-function)
                           buttons)
                 buttons)))
           (interface (make-instance 'confirmer
                         :title title
                         :ok-check ok-check
                         'ok-button ok-button
                         ;:create-callback (if ok-check '%ok-check-confirmer nil)
                         :help-key help-key	; no tooltip for top-level - do not listify
			   ;(if (or (stringp help-key) (consp help-key)) (list help-key))
                         :help-callback help-callback
                         #-lispworks4.3 #-lispworks4.3
                         :window-styles (if help-key
                                            (pushnew :contexthelp window-styles)
                                            window-styles)
                         :layout
                          (make-instance 'capi:column-layout
                           :description
                           `(,@(if (and icon message)
                                   `(,(make-instance 'capi:row-layout
                                       :description `(,(make-instance 'icon-pane :icon icon
                                                                      :accepts-focus-p nil)
                                                      ,message)
                                       :adjust :center  :gap +big-h-gap+))
                                   (listify message))
                             ,@(listify pane)
                             ,(make-instance 'capi:row-layout :description buttons
                                             :gap +h-gap+))
                           :adjust :center
                           :internal-border +padding+
                           :gap +big-v-gap+)))
           (owner (if (eq owner t) (current-top-level-interface) owner)) )
      (when icon
        (beep icon owner))
      (cond ((not owner))
            ((object-handle owner)					; already displayed
             (case (capi:top-level-interface-display-state owner)
               ((:hidden :iconic)
                (setf (capi:top-level-interface-display-state owner) :restore)))
             (when (and flash (neq owner (current-top-level-interface)))
               (flash-window owner)))
            (t (mp:process-wait 'confirmer 'object-handle owner)))	; not displayed yet
      (apply #'capi:display-dialog interface :owner owner
             (nconc (if focus (list :focus focus) ())
                    (remove-properties args
                                       '(:title :message-args
                                         :icon :owner :flash :focus :window-styles
                                         :buttons :default-button
                                         :value-function :ok-check ;:ok-function
                                         :help-key :help-callback :help-function)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  BEEP  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun beep (&optional icon pane)
 ;;; Beep accompanied by displaying text in the interface message area.
  ;; NB: capi:beep-pane does not work for non-top-level interfaces!
  #+win32
  (cond ((symbolp icon)
         (win32:message-beep (case icon
                               (:info		 win32:MB_ICONINFORMATION)
                               (:query		 win32:MB_ICONQUESTION)
                               (:error		 win32:MB_ICONERROR)
                               ((:warn :warning) win32:MB_ICONWARNING)
		                   ;(:exclamation win32:mb_iconexclamation) ; = :warn
                		   ;(:asterisk    win32:mb_iconasterisk)    ; = :info
		                   ;(:stop	  win32:mb_iconstop)	      ; = :error
                               (otherwise	 0))))
        ((stringp icon)					; message area text
         (capi:beep-pane (capi:top-level-interface pane))
         (message-area-format pane icon))
        ((consp icon)					; (control . format-arguments)
         (capi:beep-pane (capi:top-level-interface pane))
         (apply #'message-area-format pane icon))
        ((capi:beep-pane)))
  #-win32
  (progn (capi:beep-pane pane)
    (cond ((stringp icon)				; message area text
           (message-area-format pane icon))
          ((consp icon)					; (control . format-arguments)
           (apply #'message-area-format pane icon))))
  nil)

#+win32
(ffc:define-win32api |GetCurrentThreadId| () win32:DWORD) ; :module "kernel32"

;; Enumerates all nonchild windows associated with a thread by passing the handle 
;; to each window, in turn, to an application-defined callback.
#+win32
(ffc:define-win32api |EnumThreadWindows|
                     ((thread-id win32:DWORD) (wndenumproc :pointer) (lparam win32:DWORD))
                     win32:BOOL
                     :module "user32")
#+win32
(defparameter %end-dialog% (fli:make-pointer :symbol-name "EndDialog" :module "user32"))
;#+win32
;(defun register-for-end-dialog ()
;  (setf (getf (mp:process-plist mp:*current-process*) '|GetCurrentThreadId|)
;        (|GetCurrentThreadId|)))
#+win32
(defun end-dialog (thread-id &optional (value 0))
 ;;; Abort the dialog returning the dialog-specific value (MSDN for details).
  ;(when-let (thread-id (getf (mp:process-plist process) '|GetCurrentThreadId|))
  (|EnumThreadWindows| thread-id %end-dialog% value))


(defconstant +button-alist+
  `((:ok     . #+win32 ,win32:IDOK	#-win32 nil)		; "Standard" buttons
    (:cancel . #+win32 ,win32:IDCANCEL	#-win32 nil)
    ;(:abort . #+win32 ,win32:IDABORT	#-win32 nil)
    (:try    . #+win32 ,win32:IDRETRY	#-win32 nil)
    (:ignore . #+win32 ,win32:IDIGNORE	#-win32 nil)
    (:yes    . #+win32 ,win32:IDYES	#-win32 nil)
    (:no     . #+win32 ,win32:IDNO	#-win32 nil)))
    ;(:help   . #+win32 ,win32:IDHELP	#-win32 nil)))

(defun beep-box (format-control
                 &key format-arguments
                      (icon :warn) (owner t) (flash t)
                      (title (princ-to-string
                              (or (object-title (if (eq owner t)
                                                    (current-top-level-interface)
                                                    owner))
                                  (case icon
                                    (:query           #L"Confirm operation")
                                    (:error           #L"Error")
                                    ((:warn :warning) #L"Warning")
                                    (otherwise        #L"Notification")))))
                      (buttons :ok) default-button
                      timeout
                      check
                      help-key)
 ;;; Args: owner    T means an active window.
  ;;	   buttons  eyword or list of keywords and/or descriptions of the form
  ;;			(<keyword> . <plist>),
  ;;		    e.g. (:all :mnemonic-text "&Yes for all"
  ;;				:help-key "Proceed without prompting")
  ;;	   default-button  Keyword or NIL (meaning the first one is default).
  ;;       check           If true, display an additional check button centered above
  ;;			   the buttons. If the argument is a string, it specifies the
  ;;			   text of the check box, otherwise the following label is used
  ;;			   "Do not show this message any longer".
  ;;	   help-key        Help context (not tooltip),i.e. string is interpreted as anchor.
  ;;			   If true, equips the dialog with the '?' and the Help button.
  ;;                       Used to compute the help-key passed to help-callback;
  ;; Values: 1> if only standard buttons (and no listp description) are given,
  ;;		either true (means :ok, :yes, :try) or NIL (means :no, :continue);
  ;;		otherwise, if non-standard is specified,
  ;;		either the button keyword is returned or NIL if canceled.
  ;;         2> cancelp i.e. (not successp), true means the :cancel or :abort was pressed.
  ;;	     3> True if the check box was ticked.
  ;; Unfortunately, capi:popup-confirmer
  ;; 1) Does not allow to set :cancel as a default button,
  ;; 2) Does not provide localized button text with appropriate mnemonics.
  (let ((descriptions (listify buttons)))
   (if #+win32 (and (every (lambda (description)
                             (and (atom description)
                                  (assoc description +button-alist+ :test #'eq)))
                           descriptions)
                    (not (or check help-key)))
       #-win32 nil
     ;; Windows: only standard buttons are present - redirect to win32:message-box
     #+win32
     (flet ((button-member (elt) (memq elt descriptions)))
      (let* ((owner (if (eq owner t) (setq owner (current-top-level-interface)) owner))
             (value
              (progn
                (when timeout
                  (let ((thread-id (|GetCurrentThreadId|)) ;(current mp:*current-process*)
                        (default (or (cdr-assoq default-button +button-alist+) 0)))
                    (mp:process-run-function "Waiting to dismiss dialog" ()
                                             (lambda () (sleep timeout)
                                               (end-dialog thread-id default)))))
                    ;(mp:schedule-timer-relative		; timer version hangs on
                    ; (mp:make-timer 'end-dialog thread-id default) timeout)))
                (cond ((not owner))
                      ((object-handle owner)				; already displayed
                       (case (capi:top-level-interface-display-state owner)
                         ((:hidden :iconic)
                          (setf (capi:top-level-interface-display-state owner) :restore)))
                       (when (and flash (neq owner (current-top-level-interface)))
                         (flash-window owner)))
                      (t (mp:process-wait 'beep-box 'object-handle owner)))	; not yet
                (popup-before-activate owner :modal)
                (win32:message-box
                 (object-handle owner)
                 (if format-arguments
                     (apply #'format nil format-control format-arguments)
                     format-control)
                 title
                 (+ (case icon
                      (:info		win32:MB_ICONINFORMATION)
                      (:query		win32:MB_ICONQUESTION)
                      (:error		win32:MB_ICONERROR)
                      ((:warn :warning)	win32:MB_ICONWARNING)
                      ;(:exclamation	win32:mb_iconexclamation) ; = :warn
                      ;(:asterisk  	win32:mb_iconasterisk)    ; = :info
                      ;(:stop	  	win32:mb_iconstop)	      ; = :error
                      (otherwise	(or icon 0)))
                    (cond ((every #'button-member '(:ok :cancel))
                           win32:MB_OKCANCEL)
                          ((every #'button-member '(:yes :no :cancel))
                           win32:MB_YESNOCANCEL)
                          ((every #'button-member '(:yes :no))
                           win32:MB_YESNO)
                          ((every #'button-member '(:try :cancel))
                           (if (button-member :ignore)
                               2	; MB_ABORTRETRYIGNORE; MB_CANCELTRYCONTINUE!
                               5))	; MB_RETRYCANCEL
                          (t win32:MB_OK))
                    ;; When the user clicks the Help button or presses F1,
                    ;; the system sends a WM_HELP message to the owner.
                    ;; But this does not go through the help-callback of the interface!
                    ;(if help-key
                    ;    #x4000 		; MB_HELP
                    ;    0)
                    (case (position default-button descriptions)
                      (0 win32:MB_DEFBUTTON1)
                      (1 win32:MB_DEFBUTTON2)
                      (2 512)
                      (otherwise 0)))))))
        ;(modal-before-deactivate owner :modal)			; is a bit late here
        (values (member value '#.(list win32:IDOK win32:IDYES win32:IDRETRY)) ;IDTRYAGAIN!
                (member value '#.(list win32:IDCANCEL win32:IDABORT)))))
     #-win32
     nil							; no way to get here

     ;; Non-Windows or non-standard buttons are present: custom confirmer dialog
     ;; TODO: Implement timeout
     (let ((check (when check
                    (make-instance 'capi:check-button
                          :mnemonic-text (if (stringp check)
                                             check
                                             #L"&Do not show this dialog any longer")))))
       (multiple-value-bind (value successp)
           (confirm check (if format-arguments
                              (apply #'format nil format-control format-arguments)
                              format-control)
                    :title title :icon icon :owner owner :flash flash
                    :focus nil
                    :buttons descriptions :default-button default-button
                    #|:help-function (when help-key
                                     (lambda (interface)
                                       #1=(help-callback interface nil :help ;(or owner?)
                                                         (capi:help-key interface))))|#
                    #|:help-callback
                    (if help-key
                        (lambda (interface pane type help-key)
                          (if (and (eq type :help)
                                   (typep pane 'capi:push-button)) ;(memq pane buttons)
                              (if (or (eq (capi:item-data pane) :help)
                                      (eq (capi:help-key pane) :default))
                                  ;; Caller's help
                                  (help-callback interface nil :help
                                                 (capi:help-key interface))
                                  ;; For buttons context help is invoked with
                                  ;; help-key equals to data
                                  (help-callback interface pane type (capi:help-key pane)))
                              (help-callback interface pane type help-key)))
                        'help-callback)|#
                    :help-key help-key)
         ;; It is vague, whether #<capi:destroy :before (image-pane)> is invoked for dialog
         (values value
                 (not successp)
                 (when (and successp check) (capi:button-selected check)))))
) ))

(defun confirm-exit-application (&optional (owner t) (check t)
                                           (title #+ys-product yl:*thys-product-name*))
  (if (or (not check) (yl:thys-preference 'confirm-exit-application))
      (multiple-value-bind (result cancel check)
          (beep-box #L"Do you really want to exit?" ;:format-arguments (list title)
                    :icon :warn  :owner owner  :title title
                    :check check
                    :buttons '(:ok :cancel) :default-button :cancel)
        (declare (ignore cancel))
        (when check
          (setf (yl:thys-preference 'confirm-exit-application) nil))
        result)
      t))


;;; WORKAROUND: Guarantee the reasonable confirmer message width and height
;;;             due to :text-width does not work for string with linefeed embedded.
(defmethod CAPI::ENSURE-PANE-STRING :after ((pane capi:title-pane) (string string))
  (when (eq (capi:capi-object-name pane) 'CAPI::CONFIRMER-MESSAGE)
    (let ((height (max-height-in-characters string)))
      (cond ((< 1 height)
             (capi:set-geometric-hint pane :visible-min-width 
               `(min (:character ,(+ (max-width-in-characters string) 15))
                     #1=(- :screen-width (:character 15))))
             (capi:set-geometric-hint pane :visible-min-height
                      `(min (:character ,(1+ height)) :screen-height)))
            (t
             (capi:set-geometric-hint pane :visible-min-width '(min :text-width #1#))))
      (capi:set-geometric-hint pane :visible-max-width '#1#))))


#|(defun confirmer-on-create-select-text (interface)
 ;;; Select the entire :initial-value text when calling
  ;; (capi:prompt-for-string ... :popup-args '(:create-callback #))
  ;; Does not work on LWW4.6!
  (let ((pane (capi:confirmer-pane interface)))
    (capi:set-text-input-pane-selection pane 0 (length (capi:text-input-pane-text pane)))))|#

(defun prompt-for-number (message
                          &key min max initial-value ok-check pane-args popup-args
                               help-key)
 ;;; Prompter invoking locale-specific number parser
  ;; Values: result, successp
  (let ((pane (apply #'make-instance 'capi:text-input-pane
                     :title message  :title-position :top
                             ;:title-args '(:visible-min-width (:character 30))
                     :text (if initial-value (yl:f-format-to-string initial-value) "")
                     :change-callback-type :interface
                     :change-callback :redisplay-interface
                     :help-key (if help-key (getf pane-args :help-key :default))
                     (remove-properties pane-args '(:help-key))))
        (window-styles (getf popup-args :window-styles)))
    (when help-key
      (pushnew :contexthelp window-styles :test #'eq))
    (apply #'capi:popup-confirmer pane nil
	   :create-callback 'confirmer-create-callback
           :value-function (lambda (pane) (ignore-errors (values
                                            (yl:parse-number
                                             (capi:text-input-pane-text pane)))))
           :ok-check (lambda (value)
                       (and (numberp value)
                            (if ok-check (funcall ok-check value) t)
                            (if min (<= min value) t)
                            (if max (<= value max) t)))
           :cancel-button #L"Cancel"
           :help-button (if (and help-key *show-help-button*) #L"Help" nil)
           :help-function (if help-key 'help-function nil)
           :help-callback (if help-key 'help-callback nil)
           :help-key (if (or (stringp help-key) (consp help-key))
                         (list help-key)			; no tooltip
                         help-key)
           :window-styles window-styles
           (remove-properties popup-args '(:window-styles)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PROMPT-FOR-RESTART  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun prompt-for-restart (condition &key owner title
                                     ((:quit allow-quit) yl::%allow-quit%)
                                     ((:debug allow-debug) yl::%allow-debug%)
                                     (help-key (if (typep condition 'yl:helped-condition)
                                                   (yl:help-key condition)
                                                   yl::%help-key%)))
 ;;; Compute available restarts, ask the user, and invoke the chosen one.
  ;; Args: owner	 Can be run in a process different from mp:*current-process*
  ;;	   quit True	 Allow the quit restart among choice items.
  ;;		False	 Does not allow.
  ;;	 	:default Equivalent to nil if some interfaces are running in the process.
  ;;	   debug	 If true, Open Edtion invokes dbg:bug-report-form, which
  ;;			 cannot work on the highest delivery levels.
  ;;	   help-key      Help context designator (not tooltip), 
  ;;			 i.e. string is interpreted as anchor.
  ;;			 If true, equips the dialog with the '?' and the Help button.
  ;; NB1: If only return along with, possibly, continue and quit reastarts are in
  ;;	  the stack, we pop-up a simplified modal dialog instead of the restart list.
  ;; NB2: Conditions seems to be printed in long form by means of princ, not prin1!
  ;;
  ;; Remove dangerous restarts from the list of possible continuations
  (when *erroneous-pane*
    (switch-activate-pane *erroneous-pane*))
  (let* ((restarts (remove 'editor::editor-abort (compute-restarts condition)
                           :key 'restart-name))
         (quit (find-if (lambda (restart)
                          (and (eq (restart-name restart) 'abort)
                                 (stringp (CONDITIONS::RESTART-%REPORT-FUNCTION restart))
                                 (search "Quit process"
                                         (CONDITIONS::RESTART-%REPORT-FUNCTION restart)
                                         :test #'string=)))
                          restarts))
         (return (find-if (lambda (restart)				; Cancel gesture
                            (and (eq (restart-name restart) 'abort)
                                 (stringp (CONDITIONS::RESTART-%REPORT-FUNCTION restart))
                                 (search "Return to event loop"
                                         (CONDITIONS::RESTART-%REPORT-FUNCTION restart)
                                         :test #'string=)))
                          restarts))
         (continue (find 'continue restarts :key #'restart-name))	; default gesture
         (first (first restarts))
         (restarts (cond ((null quit)
                          restarts)
                         ((or (null allow-quit)
                              (and (eq allow-quit :default)		; inspect process
                                   (getf (mp:process-plist mp:*current-process*)
                                         'CAPI::DEPENDENT-INTERFACES)))
                          (prog1 (remove quit restarts) (setq quit nil)))
                         (t restarts)))
         #+Russian quit-string
         (icon (if (slot-exists-p condition 'icon)
                   (slot-value condition 'icon)
                   :error))
         (restart							; ask for restart
          (if (and (not allow-debug)				;(sys:featurep :debug)
                   (or (and (null (cdr restarts)) (eq first return))	; length=1
                       (and (eq first continue)				; length=2
                            (null (cddr restarts))
                            (eq (second restarts) return))))
              (if (beep-box #L"~@[Pane \"~a\":~2%~]~a~@[~2%OK -- ~a~]"
                            :format-arguments (list (object-title *erroneous-pane*)
                                                    condition continue)
                            :title (cond (title)
                                         (owner (capi:interface-title owner))
                                         (t #L"Error"))		; as the last resort
                            :buttons (if continue
                                         (if help-key
                                             `((:ok :help-key ,(princ-to-string continue))
                                               :cancel)
                                             '(:ok :cancel))
                                         :cancel)
                            :help-key #4=(if (eq help-key :type)
                                             (type-of condition)
                                             help-key)
                            :icon icon  :owner owner)
                  (or continue return)
                  return)
              (let ((pane (make-instance 'capi:list-panel
                            :items
                             #-Russian restarts
                             #+Russian (progn
                                         (when return
                                           (setf (first (member return restarts))
                                                 #2="  ."))
                                         (when quit
                                           (setf quit-string
                                                 (format nil
                       "~:[ ~; ~] ~s,   ."
                                             (getf (mp:process-plist mp:*current-process*)
                                                   'CAPI::DEPENDENT-INTERFACES)
                                             (mp:process-name mp:*current-process*))
                                                       (first (member quit restarts))
                                                       quit-string))
                                         restarts)
                             :selected-item continue			; DI 2010-Feb-16 ?
                             :visible-min-width '(min :text-width
                                                      (- :screen-width (:character 10)))
                             :visible-max-width nil ;'(- :screen-width (:character 20))
                             :visible-min-height '(min :text-height
                                                       (- :screen-height (:character 7)))
                             :action-callback (lambda (pane)
                                                (capi:exit-dialog
                                                 (capi:choice-selected-item pane)))
                             :callback-type '(:collection)
                             :help-key
                                #-Russian "Possible contunuations from this error"
                                #+Russian "    ")))
                (confirm pane (format nil #L"~@[Pane \"~a\":~2%~]~a"
                                      (object-title *erroneous-pane*) condition)
                         :title (or title
                                    (format nil #L"Select a restart~@[ - ~a~]"
                                            (and owner (capi:interface-title owner))))
                         :icon icon  :owner owner
                         :buttons
                         `((:ok :callback ,(lambda (&rest dummy)
                                             (declare (ignore dummy))
                                             (capi:exit-dialog
                                              (capi:choice-selected-item pane))))
                           :cancel
                           #+debug
                           (:debug :text "Debug")
                           #-debug		; image/source code delivery
                           ,.(when allow-debug
                               `((:debug :text #L"Report bug..."
                                  :callback
                                  ,(lambda (&rest dummy) (declare (ignore dummy))
                                     #-ys-product
                                     (when-let (pathname
                                                (capi:prompt-for-file #3="Bug report"
                                                 :pathname "bug-report.txt"
                                                 :operation :save))
                                       (with-open-file (*debug-io* pathname
                                                                 :direction :output
                                                                 :if-does-not-exist :create
                                                                 :if-exists :rename)
                                         (dbg:with-debugger-stack ()
                                           (dbg:bug-report-form '(:b) #3#))))
                                            ;:impact "Broken" :urgency "ASAP" :topics ?
                                     #+ys-product
                                     (bug-report-dialog condition))))))
                         :help-key #4#)))))
                                         
                    #|(buttons `(,(make-instance 'capi:push-button :text #L"OK"
                                  :data :ok :name :ok  :default-p t
                                  :visible-min-width #5='(:character 10)
                                  :callback (lambda (&rest dummy)
                                              (declare (ignore dummy))
                                              (capi:exit-dialog
                                               (capi:choice-selected-item pane)))
                                  :help-key :default)
                                ,(make-instance 'capi:push-button :text #L"Cancel"
                                  :data :cancel  :name :cancel  :cancel-p t
                                  :visible-min-width #5#
                                  :callback 'capi:abort-dialog
                                  :help-key :default)
                                #+debug
                                ,(make-instance 'capi:push-button :text "Debug"
                                  :visible-min-width #5#
                                  :callback (lambda (&rest dummy)
                                              (declare (ignore dummy))
                                              (capi:exit-dialog :debug)))
                                #-debug			; image/source code delivery
                                ,.(when allow-debug
                                   `(,(make-instance 'capi:push-button
                                       :text #L"Report bug..."
                                       :callback
                                        (lambda (&rest dummy) (declare (ignore dummy))
                                          #-ys-product
                                          (when-let (pathname
                                                     (capi:prompt-for-file #3="Bug report"
                                                          :pathname "bug-report.txt"
                                                          :operation :save))
                                            (with-open-file (*debug-io* pathname
                                                             :direction :output
                                                             :if-does-not-exist :create
                                                             :if-exists :rename)
                                              (dbg:with-debugger-stack ()
                                                (dbg:bug-report-form '(:b) #3#))))
                                            ;:impact "Broken" :urgency "ASAP" :topics ??
                                          #+ys-product
                                          (bug-report-dialog condition)))))
                                ,.(when (and help-key *show-help-button*)
                                    `(,(make-instance 'capi:push-button :text #L"Help"
                                        :mnemonic #-Russian 0 #+Russian 4
                                        :data :help
                                        :visible-min-width #5#
                                        :callback 'help-function
                                        :callback-type :interface
                                        :help-key t)))
                                ))
                     (interface (make-instance 'confirmer
                         :title (or title
                                    (format nil #L"~:[Select a restart~*~;Error : ~a~]"
                                            owner #1#))
                         :owner owner
                         :help-key (if (or (stringp help-key) (consp help-key))
                                       (list help-key)		; no tooltip
                                       #4#)
                         :help-callback (if help-key 'help-callback nil)
                         #-lispworks4.3 #-lispworks4.3
                         :window-styles (cons :resizable
                                              (if help-key '(:contexthelp) ()))
                         :layout
                          (make-instance 'capi:column-layout
                           :description
                           `(,(make-instance 'capi:row-layout
                               :description
                               `(,(make-instance 'icon-pane :icon :error)
                                 #|,(make-instance 'capi:toolbar
                                   :image-width 32 :image-height 32
                                   :items `(,(make-instance 'capi:toolbar-button
                                              :image :error
                                              :callback (if help-key 'help-function)
                                              :callback-type :interface)))|#
                                 ,(make-instance 'capi:title-pane
				   :name 'CAPI::CONFIRMER-MESSAGE ; for ENSURE-PANE-STRING
                                   ;:visible-max-height nil
                                   :text (format nil #L"~@[Pane \"~a\":~2%~]~a"
                                                 (object-title *erroneous-pane*)
                                                 condition)))
                               :adjust :center  :gap +big-h-gap+)
                             ,pane
                             ,(make-instance 'capi:row-layout
                               :description buttons
                               :gap +big-h-gap+))
                           :adjust :center  :gap +big-v-gap+))) )
                (beep :error owner)
                (capi:display-dialog interface :owner owner :focus pane) ))))|#
              #|(progn
                (beep :error)
                (capi:prompt-with-list
                 #-Russian restarts
                 #+Russian (progn
                             (when return
                               (setf (first (member return restarts))
                                     #2="  ."))
                             (when quit
                               (setf quit-string
                                      (format nil
                       "~:[ ~; ~] ~s,   ."
                                             (getf (mp:process-plist mp:*current-process*)
                                                   'CAPI::DEPENDENT-INTERFACES)
                                             (mp:process-name mp:*current-process*))
                                     (first (member quit restarts)) quit-string))
                             restarts)
                 (format nil #L"~@[Pane \"~a\":~2%~]~a"
                         (object-title *erroneous-pane*) condition)
                     ;(princ-to-string condition)
                         ;(capi:wrap-text (princ-to-string condition) 80)
                         ;(substitute #\Space #\Newline (princ-to-string condition))
                 :pane-args '(:visible-min-width (min :text-width
                                                      (- :screen-width (:character 20)))
                              :visible-max-width t
                              :help-key
                                #-Russian "Possible contunuations from this error"
                                #+Russian "    ")
                 :popup-args `(,:title ,(or title (format nil
                                                    #L"~:[Select a restart~*~;Error : ~a~]"
                                                   owner #1#))
                               ,:cancel-button ,#L"Cancel"
                               :help-button ,(if (and help-key *show-help-button*)
                                                 #L"Help"
                                                 nil)
                               :help-function ,(if help-key 'help-function nil)
                               :help-key ,(if (or (stringp help-key) (consp help-key))
                                              (list help-key)		; no tooltip
                                              #4#)
                               :help-callback ,(if help-key 'help-callback nil)
                               ,:owner ,owner
                               #-lispworks4.3 #-lispworks4.3
                               ,:window-styles (:resizable
                                                ,@(when help-key
                                                    '(:contexthelp)))
                               #+debug
                               ,@(list :apply-button "Debug"
                                       :apply-function (lambda (&rest dummy)
                                                         (declare (ignore dummy))
                                                         (capi:exit-dialog :debug)))
                               #|-debug(if (or (sys:featurep :debug)
                                         #+ys-product (null (yl:delivery-level))
                                         #+ys-product (minusp (yl:delivery-level)))
                                     (list :apply-button "Debug"
                                           :apply-function
                                           (lambda (&rest dummy) (declare(ignore dummy))
                                             (capi:exit-dialog :debug)))|#
                               #-debug			; image/source code delivery
                               ,@(when allow-debug
                                   `(,:apply-button ,#L"Report bug..."
                                     ,:apply-function
                                     ,(lambda (&rest dummy) (declare (ignore dummy))
                                        #-ys-product
                                        (when-let (pathname
                                                   (capi:prompt-for-file #3="Bug report"
                                                        :pathname "bug-report.txt"
                                                        :operation :save))
                                          (with-open-file (*debug-io* pathname
                                                           :direction :output
                                                           :if-does-not-exist :create
                                                           :if-exists :rename)
                                            (dbg:with-debugger-stack ()
                                              (dbg:bug-report-form '(:b) #3#))))
                                            ;:impact "Broken" :urgency "ASAP" :topics ??
                                        #+ys-product
                                        (bug-report-dialog condition)))) )))|#
    ;(when *erroneous-pane*
    ;  (switch-activate-pane *erroneous-pane*))
    #+Russian
    (cond ((eq restart #2#) (setq restart return))
          ((eq restart quit-string) (setq restart quit)))
    (cond #+debug ((eq restart :debug)
                   (invoke-debugger condition))
          (restart
           (case (restart-name restart)
             (abort       (if (and continue owner (eq owner (CAPI::CURRENT-DIALOG)))
                              (continue condition) ; in order not to quit current dialog
                              (invoke-restart restart)))
             (store-value (multiple-value-bind (value successp)
                              (capi:prompt-for-form #L"Input some value")
                             (if successp
                                 (store-value value condition)
                                 (abort condition))))
             (otherwise   (invoke-restart restart))))
          ((and continue owner (eq owner (CAPI::CURRENT-DIALOG)))
           (continue condition))		; in order not to quit current dialog
          (t (abort condition)) )))		; Cancel -> abort restart

(defmacro display-errors (&body body)
 ;;; Small improvement over capi:display-errors - in case of error returns nil, condition.
 `(handler-bind ((error #'prompt-for-restart)) ,@body))
 ;`(handler-case (progn ,@body)
 ;   (error (condition) (beep-box (princ-to-string condition) :icon :error)
 ;                      (values nil condition))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  BULK  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initially bound around a bulk operation and can be modified in a confirmer.
;;  T		Bound around the operation if it is not cancelable.
;;  :cancelable Bound and means that the rest of the operation is cancelable, 
;;		non-local exit through the cancel restart (catch or abort restart?).
;;  :all	Assigned in confirmer and means yes to all,
;;  :none	Assigned in confirmer and means no to all.
(defvar *in-bulk* nil)

(defmacro with-bulk ((&optional cancelable &rest format-arguments) &body body)
  (if cancelable				; => (1) NIL, (2) T if restart invoked
     `(with-simple-restart
          (cancel ,@(if (stringp cancelable)
                        `(,cancelable ,@format-arguments)
                        `(,#L"(Cancel) Do not perform bulk operation for the rest of
        the objects or records.")))
        (let ((*in-bulk* :cancelable))
          ,@body))
     `(let ((*in-bulk* t))
        ,@body)))

(defun beep-bulk (format-control &rest args &key (all t) (no t) (none no) buttons
                                            &allow-other-keys)
 ;;; Ask for confirmation with Yes, No, maybe Cancel, and when *in-bulk* is true,
  ;; additionalal buttons like "Yes to all" or "No to all".
  ;; Args: all, none Mnemonic string, plist, or true means displaying the
  ;;		     corresponding button.
  ;;		     NIL means that the button is not needed at all.
  ;;	   args      Key parameters accepted by beep-box except for :buttons.
  ;; Values: 1) True or false
  ;;         2) True if canceled
  ;;      or non-local exit to the cancel restart.
  ;; Usage:
  ;;	(case *in-bulk*
  ;;	  (:all t)
  ;;	  (:none nil)
  ;;	  (t (beep-bulk ...)))
  (let ((buttons `(,@(when no
                       '(:no))
                   ,@(when (and *in-bulk* all)
                       `(,(cond ((consp all) `(:all ,@all))
                                ((stringp all) `(:all :mnemonic-text ,all))
                                (:all))))
                                        ;:help-key "Proceed without prompting")
                   ,@(when (and *in-bulk* none)
                       `(,(cond ((consp none) `(:none ,@none))
                                ((stringp none) `(:none :mnemonic-text ,none))
                                (:none))))
                   ,@(when (or (eq *in-bulk* :cancelable)	;(or (not *in-bulk*)
                               (memq :cancel buttons))
                       `(,:cancel)))))
    (multiple-value-bind (result cancel)
        (apply #'beep-box
               (if (eq *in-bulk* :cancelable)			; append extra explanation
                   (if-bind (restart (find-restart 'cancel))	; for the cancel restart
                     (let* ((text (princ-to-string restart))
                            (pattern (load-time-value (string-append #\( #L"Cancel" #\))))
                            (position (mismatch pattern text)))
                       (string-append format-control
                                      #\Newline #\Newline #L"Cancel" " - "
                                      (if (iplusp position) (subseq text position) text)))
                     format-control)
                   format-control)
               :buttons (cons (if (or (rest buttons)		; two buttons at least
                                      (memq (first-or-self (first buttons))
                                            '(:no :all :none)))
                                  :yes
                                  :ok)				; empty or :cancel only
                              buttons)
               ;:default-button (first buttons) 
               (sans args :buttons :all :no :none))
      (if cancel
          (if (eq *in-bulk* :cancelable)
              (if-bind (restart (find-restart 'cancel))
                (invoke-restart restart)
                (values nil cancel))
              (values nil cancel))
          (case result
            ((:all :none) (eq (setq *in-bulk* result) :all))
            (:yes         t)
            (:no          nil)
            (otherwise    result))))))
