;;;-*- Mode: Lisp; Package: COMMON-LISP-USER -*-
;;; Ystok-Widgets - LispWorks CAPI-based GUI framework
;;; Copyright (c) 2003-2010 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Inline editor for pinboard objects
;;;
;;; DISPLAYER	 - an abstract object, which is not passed to any CAPI-specific operation
;;;		   explicitly. (Even not to capi:pinboard-object-pinboard!)
;;; EDITING-PANE - either inline simple pane, maybe within overlaying interface
;;;		 - or popup editing device.
;;; NB:  pinboard   win32:ws_tabstop,        .
;;;     #\Tab

(in-package :ystok.widgets)

(defgeneric editing-pane-type (displayer)
 (:method (displayer) nil))

(defgeneric pane-args (displayer)
 (:method (displayer) ())
 (:method ((displayer capi:capi-object)) (capi:capi-object-property displayer :pane-args))
)

(defgeneric (setf pane-args) (value displayer)
 (:method (value (displayer capi:capi-object))
  (setf (capi:capi-object-property displayer :pane-args) value))
)

;; Dynamic scope bound to current pinboard guaranteed for the below three generics.
(defvar *chooser*)

(defgeneric displayer-rectangle (displayer &optional subscript screen)
 ;;; Value: External coordinates (if screen is true, relative to screen) and size
 (:method ((pane capi:simple-pane) &optional subscript screen)
  ;; NB: capi:simple-pane-visible-size returns view-port size without external 3D-boarder
  (declare (ignore subscript))
  (capi:with-geometry pane
    (if screen
        (multiple-value-bind (left top) (capi:convert-relative-position pane
                                         (capi:element-screen pane) 0 0)
          (if (capi:simple-pane-visible-border pane)
              (make-rect (- left 2) (- top 2) capi:%width% capi:%height%)
              (make-rect left top capi:%width% capi:%height%)))
        (make-rect capi:%x% capi:%y% capi:%width% capi:%height%)))))

(defgeneric read-displayer-data (displayer &optional subscript))
(defgeneric write-displayer-data (value displayer &optional subscript))
;(defgeneric test-function (displayer))		; for editing pane type like choice

(defgeneric current-displayer (chooser)
 (:method (chooser) nil)
 (:method ((chooser CAPI::SIMPLE-ELEMENT)) chooser) )

(defgeneric current-subscript (chooser)
 (:method (chooser) nil)
 (:method ((chooser capi:capi-object)) (capi:capi-object-property chooser :realm)) )

(defgeneric own (pane)
 ;;; True if the pane is linked to single displayer and not shared by others
  ;; Can be either a list of pane types or T.
 (:method (pane) nil))

(defclass editing-pane-mixin ()
 (;; Current or previous displayer being edited or, for own pane, linked to;
  ;; tracking the previous is for choices not to reinitialize collection items.
  (current-displayer :accessor current-displayer  :initarg :displayer  :initform nil)
  (own :accessor own  :initarg :own  :initform nil)))

(defgeneric make-editing-pane (type pinboard displayer)
 ;;; Can be called only to create a pane dynamically, after the pinboard has been created!
 (:method (type pinboard displayer)
  ;; Set the own slot if the pane-args mentions it.
  ;; Args: type  May not be a subtype of editing-pane-mixin!
  (let ((pane-args (pane-args displayer)))
    (apply #'make-instance type
           :callback-type :none ;:data 
           :visible-min-width nil  :visible-max-width nil
           :visible-min-height nil :visible-max-height nil
           :visible-border nil		; works for text-input-pane only
                 			;:window-styles '(:borderless) - does not help
           :x 0 :y 0  :width 0 :height 0
           :font (or (getf pane-args :font) (capi:simple-pane-font pinboard))
           (remove-properties pane-args '(:font :items :min-width))))))


;;; Mixin class for defining displayers. It should preceed other classes.

(defclass editable-pinboard-object-mixin ()
 ((editing-pane-type :reader editing-pane-type
                     :initarg :editing-pane-type
                     :initform 'capi:text-input-pane)
  (pane-args :accessor pane-args  :initarg :pane-args  :initform nil)))
  ;(test-function :reader test-function  :initarg :test-function  :initform #'equal)

;;; Pinboard is a kind of chooser

(defgeneric pinboard-font (pinboard)
 ;;; Valeu: An instance of gp:font-description or NIL.
  (:method (pinboard)
   (declare (ignore pinboard))
   nil))

(defclass editable-pinboard (capi:pinboard-layout)
 (;; Type of inline editing interface: t or plist means overlaying-interface, nil - none.
  ;; plist    -     :
  ;;	'([:default-p <object>] [:cancel-p <object>]),
  ;;  <object> ::= <capi-object> | (<selection-callback> . <callback-type>) | button 
  (wrapper :initarg :wrapper  :initform nil)
  (overlay :accessor overlay :initform nil)	; editing device: pane or interface
  (current-displayer :accessor current-displayer :initform nil)	; displayer being edited
  (current-subscript :accessor current-subscript :initform nil)	; displayer specifiers
  (editing-panes :accessor editing-panes :initform nil) 	; cached editing widgets
  (overlaying-interface :initform nil)		; cached overlaying interface
  ;; Auto-enter the editing pane just on selecting (receiving focus by) the displayer 
  ;; without additional action (like pressing Enter or double-clicking).
  (edit-on-enter :accessor edit-on-enter  :initarg :edit-on-enter  :initform nil)
  ;; Auto-save the editing pane after the displayer looses focus.
  (save-on-exit :accessor save-on-exit  :initarg :save-on-exit  :initform nil)))

(defmethod initialize-instance :after ((self editable-pinboard) &key font)
  (when (and (null font) (setq font (pinboard-font self)))
    (setf (capi:simple-pane-font self) font)))

;;; Low-level implementation insiders

(defgeneric load-editing-pane (pane pinboard displayer &optional subscript)
 ;;; Initialize editing pane, then displayer -> read-displayer-data -> pane
 (:method :after ((pane editing-pane-mixin) pinboard displayer &optional subscript)
  (setf (current-displayer pane) displayer)) )

;; Clean up on finishing
(defgeneric unload-editing-pane (pane)
 (:method (pane) (declare (ignore pane))))

;; pane -> write-displayer-data -> (current-displayer pinboard)
(defgeneric save-editing-pane (pane pinboard))

;;; capi:editor-pane as editing pane

;(defparameter *default-text-edit-font-size* 8)

(defun %editing-font-attributes (&optional font)
 ;;; Get font-attributes plist used to create font for capi:editor-pane
  ;; Args: font  Pinboard font or plist or nil
  ;; NB: The Editor only allows a fixed width font (up to LW version 5.0).
  ;; OLD: If charset argument is given, Courier is chosen, otherwise Courier New.
  (list* :size (cond ((GP::FONT-P font)
                      (gp:font-description-attribute-value (gp:font-description font)
                                                           :size))
                     ((consp font)
                      (getf font :size 8))
                     (8))	;*default-text-edit-font-size*
         :family (cond ((and (consp font) (getf font :family)))
                       ((GP::FONT-P font)
                        (gp:font-description-attribute-value (gp:font-description font)
                                                             :family))
                       (t #+lispworks4 "Courier"
                          #-lispworks4 "Helvetica"))
         :pitch (or (and (consp font) (getf font :pitch))
                    #+lispworks4 :fixed #-lispworks4 :wild)
         '(:slant :roman
           :weight :normal)))
           ;#+Russian #+Russian :charset :Russian

(defgeneric editing-pane-font (pinboard type displayer)
 ;;; Useful within specializations of make-editing-pane
 (:method ((pinboard capi:pinboard-layout) (type (eql 'capi:editor-pane)) displayer)
  (apply #'gp:make-font-description
         (%editing-font-attributes (capi:simple-pane-font pinboard)))) )

;;;;;;;;;;;;;;;;;;;;;;;;;;;  OVERLAYING-INTERFACE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Inline editing interface.
;;; The push-button-panel is needed to provide the Escape gesture.
;;; For the Enter gesture, it is enough to exploit the callback of
;;; - text-input-pane or text-input-choice
;;; - but not option-pane (neither in LWW4.3 nor LWW4.4).
;;;
;;;  :
;;;        :cancel-p/:default-p t,
;;;        overlaying-interface, .. 
;;;  buttons     Return  Escape.

(capi:define-interface overlaying-interface ()
 ((pane :initarg :pane  :initform nil))
 (:panes
  (buttons capi:push-button-panel :items 
   `(,(make-instance 'capi:push-button  :cancel-p t  :data nil
                     :selection-callback 'finish-editing  :callback-type :interface-data
                     :visible-min-height 0  :visible-max-height 0
                     :accepts-focus-p nil
                     :help-key nil)
     ,(make-instance 'capi:push-button  :default-p t  :data t
                     :selection-callback 'finish-editing  :callback-type :interface-data
                     :visible-min-height 0  :visible-max-height 0
                     :accepts-focus-p nil
                     :help-key nil))   
   :visible-min-width 0 
   :accepts-focus-p nil))
 (:layouts
  (main-layout capi:column-layout ()  :y-ratios '(1 nil)  :y-gap 0))
 (:default-initargs
  :help-key :parent))					; redirect help requests to parent

(defun finish-editing (overlaying-interface data)
  (let ((pinboard (capi:element-parent overlaying-interface)))
    (when data
      (save-editing-pane overlaying-interface pinboard))
    (reset-overlay pinboard)))

(defun ensure-overlaying-interface (pinboard)
  (if-bind (overlaying-interface (slot-value pinboard 'overlaying-interface))
    (let* ((buttons (slot-value overlaying-interface 'buttons))
           (cancel-button (capi:get-collection-item buttons 0))
           (ok-button (capi:get-collection-item buttons 1)))
      (allf 'finish-editing
        (capi:callbacks-selection-callback cancel-button)
        (capi:callbacks-selection-callback ok-button))
      (allf :interface-data
        (capi:callbacks-callback-type cancel-button)
        (capi:callbacks-callback-type ok-button))
      overlaying-interface)
    (let* ((wrapper (slot-value pinboard 'wrapper))
           (type (if (or (eq wrapper t) (consp wrapper)) 'overlaying-interface wrapper))
           (new (make-instance type
                               :visible-min-width nil :visible-max-width nil
                               :visible-min-height nil :visible-max-height nil)))
      #-lispworks6 (CAPI::FORCE-PINBOARD-ELEMENT-INITIAL-GEOMETRY new)
      (push new (capi:layout-description pinboard))
      (setf (slot-value pinboard 'overlaying-interface) new))))

(defun wrap-with-interface (overlaying-interface pane)
  (setf (capi:layout-description (slot-value overlaying-interface 'main-layout))
          (list pane (slot-value overlaying-interface 'buttons))
        (slot-value overlaying-interface 'pane) pane)
  overlaying-interface)

(defmethod save-editing-pane ((pane overlaying-interface) pinboard)
  (save-editing-pane (slot-value pane 'pane) pinboard))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  WORKHORSE FUNCTIONS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; NB: In YGrid, due to optimization of hide/show operations, reset occurs
;;;	in step-by-step manner rather than entirely:
;;;	- first, (nilf current-displayer current-subscript)
;;;	- next, maybe nilf and hide overlay

(defun ensure-editing-pane (pinboard displayer)
  (when-let (pane-type (editing-pane-type displayer))
    ;(print (pane-args displayer))
    (let* ((own (getf (pane-args displayer) :own))		; quicker computing?
           (pane (if (if (consp own)
                         (member pane-type own :test #'subtypep)
                         own)
                     (find displayer (editing-panes pinboard) :key 'current-displayer)
                     (find-if (lambda (pane) (and (not (own pane))
                                                  (eq (type-of pane) pane-type)))
                              (editing-panes pinboard)))))
      (if pane
          pane
          (let ((new (make-editing-pane pane-type pinboard displayer)))
            #-lispworks6 (CAPI::FORCE-PINBOARD-ELEMENT-INITIAL-GEOMETRY new)
            (unless (slot-value pinboard 'wrapper)
              (push new (capi:layout-description pinboard)))
            (push new (editing-panes pinboard))
            (values new t))))))

(defun enter-editing-pane (pinboard displayer &optional subscript)
 ;;; Args: pinboard   editable-pinboard
  ;;	   displayer  Pinboard (or other) object displaying some content at position
  ;;		      prescribed by the subscript.
  ;; Value: 1) An instance of editing pane on success.
  ;;        2) True if the pane was just created
  ;;	       (useful for setting caret position and other workarounds)
  ;; The function only does:
  ;; - position and resize the overlay at coordinates returned by the method
  ;;   on displayer-rectangle,
  ;; - initialize editing-pane invoking the method on reinitialize-editing-pane,
  ;; - copy the displayed value into the pane.
  (multiple-value-bind (pane created-p) (ensure-editing-pane pinboard displayer)
    (when pane
      (let ((overlay (if (slot-value pinboard 'wrapper) ;*wrap-overlay-with-interface*
                         (wrap-with-interface (ensure-overlaying-interface pinboard) pane)
                         pane)))
        (hide-editing-panes pinboard nil pane)
        (let ((*chooser* pinboard))
          (rect-bind (x y width height) (displayer-rectangle displayer subscript)
            (setf (capi:pinboard-pane-position overlay) (values x y)
                  (capi:pinboard-pane-size overlay) (values width height)))
          (load-editing-pane pane pinboard displayer subscript))	; extract value
        (setf (current-displayer pinboard) displayer
              (current-subscript pinboard) subscript
              (overlay pinboard) overlay)
        (capi:show-pane overlay)
        (capi:set-pane-focus pane))
      (values pane created-p))))

(defun hide-editing-panes (pinboard &optional reset exclude)
 ;;; Args: reset   If true, hide (or destroy on non-Windows) overlay without saving.
  ;;       exclude A pane that is about to receive focus; so we do not hide it
  ;;	 	   except when it equals to overlay and reset is true
  (dolist (pane (editing-panes pinboard))
    (unless (eq pane exclude)
      (capi:hide-pane pane)
      (unless (eql (capi:pinboard-pane-size pane) 0)
        (setf (capi:pinboard-pane-size pane) (values 0 0)
              (capi:pinboard-pane-position pane) (values 0 0)))))
  (when reset
    (reset-overlay pinboard)))

(defun reset-overlay (pinboard &optional (focus t))
 ;;; Hide or destroy overlay without saving or just after saving
  (with-slots (overlay current-displayer current-subscript
               overlaying-interface wrapper #-win32 editing-panes) pinboard
    #-win32 (when overlay
              (yl:debug-format :ywidgets "~&reset-overlay ~s" overlay)
              (capi:destroy overlay)
              (setf (capi:layout-description pinboard)
                    (delete overlay (capi:layout-description pinboard)))
              (when focus
                (capi:set-pane-focus pinboard))
              (nilf editing-panes overlaying-interface))		; clear cache
    #+win32 (when overlay
              (capi:hide-pane overlay)
              (setf (capi:pinboard-pane-size overlay) (values 0 0))
              (when focus
                (capi:set-pane-focus pinboard))
              (if (eq overlay overlaying-interface)
                  ;;       ,
                  ;;      pinboard's plist.
                  (flet ((reinitialize-button (button pair)
                         (cond ((consp pair)
                                (setf (capi:callbacks-selection-callback button)(car pair)
                                      (capi:callbacks-callback-type button) (cdr pair)))
                               (pair			; pair is a button itself
                                (setf (capi:callbacks-selection-callback button)
                                       (capi:callbacks-selection-callback pair)
                                      (capi:callbacks-callback-type button)
                                       (capi:callbacks-callback-type pair)))
                               (t
                                (setf (capi:callbacks-selection-callback button) nil)))))
                    (unload-editing-pane (slot-value overlaying-interface 'pane))
                    (let* ((buttons (slot-value overlaying-interface 'buttons))
                           (wrapper wrapper))
                      (reinitialize-button (capi:get-collection-item buttons 0)
                                        (if (consp wrapper) (getf wrapper :cancel-p)))
                      (reinitialize-button (capi:get-collection-item buttons 1)
                                        (if (consp wrapper) (getf wrapper :default-p)))))
                  (unload-editing-pane overlay)))
    (nilf overlay current-displayer current-subscript)))

(defun update-and-reset (pinboard)
  (when-let (overlay (overlay pinboard))
    (save-editing-pane overlay pinboard)
    (reset-overlay pinboard)))

(defmethod find-pane-to-focus ((pinboard editable-pinboard) &optional test)
  (let ((overlay (overlay pinboard)))
    (cond ((and overlay (enabled overlay))
           (let* ((overlaying-interface (slot-value pinboard 'overlaying-interface))
                  (pane (if (eq overlay overlaying-interface)
                            (slot-value overlaying-interface 'pane)
                            overlay)))
             (when (imp test (funcall test pane))
               pane)))
          ((and (capi:accepts-focus-p pinboard) (imp test (funcall test pinboard)))
           pinboard))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  EDITING COLLECTION  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Q: Should %cache-items% be any sequence or must be a simple vector?

(defmethod %cache-items% ((object capi:capi-object))
  (or (capi:capi-object-property object '%cache-items%) *unbound-value*))
(defmethod (setf %cache-items%) (items (object capi:capi-object))
  (setf (capi:capi-object-property object '%cache-items%) items))

(defun ensure-editing-collection-items (pane displayer subscript &optional dropped-width)
  ;; Args: pane	     If NIL, return the items rather than set assign them.
  ;;       displayer Its (getf pane-args :items) is
  ;;		      - either a function object. e.g. lambda-expression, of signature:
  ;;			 (displayer subscript),
  ;;		      - or an arbitrary form evaluated by yl:tiny-eval.
  ;;		     Both can return a second value - true means "do not cache items".
  ;;       dropped-width
  ;;		     Width hint if collection is a drop-down list or combo-box.
  ;; ASSUMPTION: *chooser* is already bound by the caller and can be evaluated by
  ;;		 the items function.
  ;; NB1: Filling cache goes through (setf capi:collection-items) -> capi:collection-items
  ;; NB2: Usually, should be put inside (capi:with-atomic-redisplay (pane) ...)
  ;; SEE ALSO: grid::%choice-column-on-character proceeds similary.
  (let ((items (if pane (capi:collection-items pane) nil))
        (%cache-items% (%cache-items% displayer)))
    (cond ((eq items %cache-items%)
           items)
          ((neq %cache-items% *unbound-value*)
           (when pane
             (setf (capi:collection-items pane) %cache-items%)
             #1=(when dropped-width
                  (set-dropped-width pane dropped-width)))
           %cache-items%)
          (t
           (multiple-value-bind (sequence dont-cache)
               (cond ((functionp (setq items (getf (pane-args displayer) :items)))
                      ;(or (symbolp items))
                      (funcall items displayer subscript))
                     ((and (consp items) (eq (car items) 'function)); forward reference
                      (funcall (cadr items) displayer subscript))   ; before definition
                     ((and (consp items) (eq (car items) 'funcall)) ; deprecated code for
                      (apply (car items) (cdr items)))		    ; back compatibility
                     (t (yl:tiny-eval items)))			    ; arg2 := displayer?
             (if pane
                 (setf (capi:collection-items pane) sequence
                       sequence (capi:collection-items pane))
                 (setq sequence (coerce sequence 'vector)))
             (unless dont-cache
               (setf (%cache-items% displayer) sequence))
             (when pane
               #1#)
             sequence)))))
           #|(setf (capi:collection-items pane) items
                 (%cache-items% displayer) (capi:collection-items pane))
           #1#))))|#

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  SPECIFIC WIDGETS  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod load-editing-pane ((pane capi:text-input-pane) pinboard displayer
                              &optional subscript)
  (setf (capi:text-input-pane-text pane) (read-displayer-data displayer subscript)
        (capi:text-input-pane-callback pane)
         (lambda (&rest dummy) (declare (ignore dummy)) (update-and-reset pinboard))))

(defmethod save-editing-pane ((pane capi:text-input-pane) *chooser*)
 ;;; Used for capi:text-input-choice as well
  (write-displayer-data (capi:text-input-pane-text pane)
                        (current-displayer *chooser*) (current-subscript *chooser*)))

(defun load-option-pane-choice-mixin (pane displayer subscript)
 ;;; Cannot turn this into a :before method since YGrid is used another code
  ;; Test-function - comparison between two items, not data; 
  ;; defaults to NIL and means "class-specific", e.g. or eql for button-panel.
  (let ((pane-args (pane-args displayer)))
    (setf (capi:collection-print-function pane) (getf pane-args :print-function
                                                      'princ-to-string)
          (capi::collection-data-function pane) (getf pane-args :data-function 'identity)
          (capi:collection-test-function pane) (getf pane-args :test-function)
          (capi:option-pane-visible-items-count pane) (getf pane-args :visible-items-count
                                                            :default) )
    (ensure-editing-collection-items pane displayer subscript
                                     (getf pane-args :min-width)) ))

(defmethod load-editing-pane ((pane capi:text-input-choice) pinboard displayer
                              &optional subscript)
  (let ((*chooser* pinboard))					; provide dynamic binding
    (load-option-pane-choice-mixin pane displayer subscript)
    (setf (capi:text-input-pane-text pane) (read-displayer-data displayer subscript)))
  (setf (capi:text-input-pane-callback pane)
        (lambda (&rest dummy) (declare (ignore dummy)) (update-and-reset pinboard))))

(defmethod load-editing-pane ((pane capi:option-pane) *chooser* displayer
                              &optional subscript)
 ;;; NB: capi:option-pane -    callback,  action-callback!
  (load-option-pane-choice-mixin pane displayer subscript)
  (setf (capi:choice-selected-item pane) (read-displayer-data displayer subscript)))

(defmethod save-editing-pane ((pane capi:option-pane) *chooser*)
  (write-displayer-data (capi:choice-selected-item pane)
                        (current-displayer *chooser*) (current-subscript *chooser*)))
