;;;-*- Mode: Lisp; -*-
;;; Ystok-Widgets - LispWorks CAPI-based GUI framework
;;; Copyright (c) 2003-2013 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Icon-related features (including non-Windows platforms)

(in-package :ystok.widgets)

;;; Bitmaps 32x32 implementing standard icons.
;;; They are used in promters to emulate gp:load-icon-image on the platforms
;;; other than Windows.
#-win32
(gp:register-image-translation :info
  #.(gp:read-external-image (lw:current-pathname "images/info.bmp")
                            :transparent-color-index 13))
#-win32
(gp:register-image-translation :query
  #.(gp:read-external-image (lw:current-pathname "images/query.bmp")
                            :transparent-color-index 13))
#-win32
(gp:register-image-translation :warn
  #.(gp:read-external-image (lw:current-pathname "images/warn.bmp")
                            :transparent-color-index 13))
#-win32
(gp:register-image-translation :error
  #.(gp:read-external-image (lw:current-pathname "images/error.bmp")
                            :transparent-color-index 13))

(defgeneric imagep (object)
 (:method (object) nil)
 (:method ((object gp:image)) t))

(defclass icon-pane (capi:output-pane)
 ((data :initarg :icon)				; the source of the image
  (image :reader image :initform nil))		; the result of gp:load-image or the like
 (:default-initargs				; - can only be set after pane is created
  :display-callback 'icon-pane-display
  :background #-(and lispworks4 win32) :color_3dface
              #+(and lispworks4 win32) 'win32:color_3dface
						; for text like "File ... does not exist"
  :foreground #-(and lispworks4 win32) :color_graytext
              #+(and lispworks4 win32) 'win32:color_graytext
  :internal-min-width 32 :internal-max-width t
  :internal-min-height 32 :internal-max-height t
  :horizontal-scroll nil :vertical-scroll nil
  :internal-border nil :visible-border nil))

(defun free-image (pane &optional makunbound)
 ;;; Free old image if any
  ;; Used by: Methods capi:destroy or interface-on-destroy
  (when-let (image (image pane))
    (when (and (imagep image) (not (gp:image-freed-p image)))
      (gp:free-image pane image)))
  (when makunbound
    (slot-makunbound pane 'data)
    (setf (slot-value pane 'image) nil)))

(defmethod (setf image) (value (pane icon-pane))
 ;;; Args: value Actually, always gp:image for icon-pane.
  ;; CAUTION: If specialized, this method would override the one on image-pane!
  (free-image pane)
  (setf (slot-value pane 'image) value)
  (gp:invalidate-rectangle pane))		; -> icon-pane-display

(defmethod (setf image) ((value gp:external-image) (pane icon-pane))
  (setf (image pane) (gp:convert-external-image pane value)))

(defmethod (setf image) ((value symbol) (pane icon-pane))
 ;;; Args: value On all platforms, a non-keyword symbol identifies an image
  ;;		 in gp:*default-image-translation-table*
  ;;		 On Windows, a keyword identifies the corresponding "standard" icon.
  (if value
      (setf (image pane)
            (cond #+win32
                  ((keywordp value)
                   (gp:load-icon-image pane (case value	; subst by whatever accpeted by gp
                                              (:info     :information)
                                              (:query    :ques)
                                              (:warn     :warning)
                                              (otherwise value))
                                       :width (gp:port-width pane)
                                       :height (gp:port-height pane)))
                  (t (gp:load-image pane value))))
      (call-next-method)))				; depends on pane subclass

(defun icon-pane-display (pane x y width height)
  (declare (ignore x y width height))
  (let ((image (image pane)))
    (cond ((imagep image)
           (gp:draw-image pane image 0 0))
          ((and (slot-boundp pane 'data) (setq image (slot-value pane 'data)))
           (setf (image pane) image)))))	; -> recurse into image-pane-display 

;;;;;;;;;;;;;;;;;;;;;;  WINDOW top-left corner and taskbar  ;;;;;;;;;;;;;;;;;;;;;;
;;; TODO: Provide code for %register-icons on non-Windows platforms
;;; BAD: In an aplication delivered with LWW4.4.6
;;;	 - :icon-small-data small icons seems not working (looks dirty, maybe stored in
;;;	   :icon-data the large is resized and displayed instead of the small).

(defparameter *icon-interface-types* ())

(defmacro define-icons (symbol &key icon-data icon-small-data
                                    (icon #+lispworks4 icon-small-data))
 ;;; Args: symbol CAPI interface class name.
  ;;       icon   NIL, small-file, or pair (small-file . large-file) of namestrings
  ;;		  to be loaded in run-time from the product installation directory.
  ;; Usage:
  ;;  (1) (define-icons interface-class-name
  ;;		        :icon-data "images/my32x32.ico"
  ;;		        :icon-small-data "images/my16x16.ico")
  ;;  (2) Call (ensure-icons interface) in :create-callback or the like.
  #+win32
  (let ((location (lw:pathname-location (or *compile-file-truename* *load-truename*))))
   `(progn
     (setf ,@(when icon
               `((get ',symbol :icon) ,icon))
           ,@(when icon-data
               `((get ',symbol :icon-data)
                 (win32:read-icon-file-data (merge-pathnames ,icon-data ,location))))
           #-lispworks4
           ,@(when icon-small-data
               `((get ',symbol :icon-small-data)
                 (win32:read-icon-file-data (merge-pathnames ,icon-small-data ,location))))
           ,@(when (or icon-data icon-small-data)
               `(*icon-interface-types* (adjoin ',symbol *icon-interface-types*))))
     ,@(when (or icon-data #-lispworks4 icon-small-data)
         `((defmethod CAPI::CAPI-INTERFACE-WINDOW-CLASS ((self ,symbol))
             (%window-class-name ',symbol))
           (%register-icons ',symbol)))	; (eval-when (:load-toplevel) ...) for debugging
     ',symbol)))

#+win32
(defun %window-class-name (symbol)
  (string-append (package-name (symbol-package symbol)) #\_ (symbol-name symbol)))

#+win32
(defun %register-icons (&optional (list *icon-interface-types*))
  (dolist (symbol (listify list))
     (WW::REGISTER-WINDOW-ATTRIBUTES (%window-class-name symbol)
               :icon-data (get symbol :icon-data)
               :icon-small-data (get symbol :icon-small-data))))

#+(and win32 (not lispworks-personal-edition))
(lw:define-action "Delivery Actions" "Keep icon interface type names"
  (lambda () (appendf (lw:delivery-value :keep-symbol-names) *icon-interface-types*)
             (appendf (lw:delivery-value :keep-symbols) *icon-interface-types*))) ; plists
#+win32
(lw:define-action "When starting image" "Register window icons" '%register-icons)

(defun %ensure-icons (interface symbol &optional (icon (get symbol :icon)))
 ;;; Helper loading icons from files during run-time
  ;; Args: symbol Whose plist stores the namestrings and handles.
  ;;	   icon   Small-file or pair (small-file . large-file) of namestring
  ;; Usage: (setf (get 'my-face :icon) "images/small.ico")
  ;;	or  (setf (get 'my-face :icon) ("images/small.ico" . "images/lagre.ico"))
  ;;	    (defmethod interface-on-create :before ((self my-face)) 
  ;;	            or capi:interface-display :after ((self my-face))
  ;;	      (ywi:ensure-icons self))
  ;; NB: To set run-time icons only in a delivered application, not while debugging in IDE,
  ;;	 put their files only into deliver/images/ but not to images/.
  #+win32
  (flet ((set-icon (small &optional large)		; send messages with icon handles
           (when-let (handle (capi:simple-pane-handle interface))
             (when (and small (not (zerop small)))
               (win32:send-message handle win32:WM_SETICON 0 small))	; ICON_SMALL
             (when (and large (not (zerop large)))
               (win32:send-message handle win32:WM_SETICON 1 large)))))	; ICON_LARGE
    (cond ((numberp icon)
           (set-icon icon))
          ((and (consp icon) (numberp (car icon)))
           (set-icon (car icon) (cdr icon)))
          (icon						; load and cache icon handle
           (let ((small (win32::load-image		; =>0 if path/file does not exists
                         0 (namestring (yl:product-pathname (first-or-self icon)))
                         1 16 16 #x10)) ;IMAGE_ICON  LR_LOADFROMFILE
                 (large (when (consp icon)
                          (win32::load-image
                           0 (namestring (yl:product-pathname (cdr icon)))
                           1 32 32 #x10))))
             (setf (get symbol :icon) (if large (cons small large) small))
             (set-icon small large))))))

(defgeneric ensure-icons (interface &key symbol)
 ;;; Load icon images from files during run-time
  ;; Args: symbol Whose plist stores the namestrings and handles.
  ;;		  May differ from the interface type for subclasses.
 (:method ((interface capi:interface) &key (symbol (type-of interface)))
  (%ensure-icons interface symbol)))
