;;;-*- Mode: Lisp; -*-
;;; Ystok-Widgets - LispWorks CAPI-based GUI framework
;;; Copyright (c) 2003-2012 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Progress meter and progress dialog

(in-package :ystok.widgets)

(defun progress-meter (pane &optional n title message interface-title)
 ;;;  Passive synchronous progress dialog or dialog-less progress indicator.
  ;;  The function either initilizes (if pane is null or n=0)
  ;;  or progress to the n'th value or by step one (if not specified).
  ;; Args: pane  NIL =create a capi:progress-bar instance, contained in a non-modal dilog;
  ;;		 or capi:progress-bar to be progressed;
  ;;             or arbitrary pane within a top-level interface having a message area;
  ;;		 or :podium keyword
  ;;       n ::= number - ending (when creating progress-bar, defaults to 100)
  ;;			  or current position (when have been already created);
  ;;		 or NIL - increase by 1;
  ;;		 or :destroy - destroy the dialog or flush the message area.
  ;;	   title For a progress-bar pane, title displayed above.
  ;;		 For message area meter, its prefix text.
  ;;	   message
  ;;		 String to be displayed below the progress-bar pane.
  ;;	   interface-title
  ;;		 Window title for a progress-bar pane dialog.
  ;;		 Usually yl:*thys-product-name* passed.
  ;; Value: Newly created or old capi:progress-bar instance, or unspecified.
  ;; NB1: Closing the dialog by the user is harmless and does cancels the operation.
  ;; NB2: In dialog-less case, the hack is to store n in the capi:interface-message-area
  ;;	  slot.
  (cond ((null pane)
         (setq pane (capi:contain (make-instance 'capi:progress-bar
                                                 :title title 
                                                 :message message
                                                 :end (or n 100)) 
                                  :title (or interface-title #L"In progress")
                                  :best-width '(/ :screen-width 2)
                                  :best-x '(/ :screen-width 4)
                                  :best-y '(/ :screen-height 3))))
        ((typep pane 'capi:progress-bar)
         (capi:apply-in-pane-process pane
           (lambda (pane n title message)
             (if (eq n :destroy)
                 (capi:destroy (capi:top-level-interface pane))
                 (progn (if (integerp n)
                            (setf (capi:range-slug-start pane) n)
                            (incf (capi:range-slug-start pane)))
                   (when title
                     (setf (capi:titled-object-title pane) title))
                   (when message
                     (when-let (message-pane (CAPI::TITLED-OBJECT-MESSAGE-PANE pane))
                       (setf (capi:titled-object-message pane) message)
                       (let ((height (max-height-in-characters message)))
                         (when (< 1 height)
                           (capi:set-geometric-hint message-pane
                            :visible-min-height `(min (:character ,height) :screen-height))
                           (capi:invalidate-pane-constraints
                            (capi:element-parent pane)))))))))
           pane n title message)
         pane)
        (t
         (let ((interface (if (eq pane :podium)
                              (capi:locate-interface
                               #+win32 'lw-tools::lispworks-podium
                               #-win32 'lw-tools:lispworks-echo-podium)
                              (capi:top-level-interface pane))))
           (when (and interface (capi:interface-message-area interface))
             (capi:execute-with-interface interface
               (lambda (interface n title)
                 (cond ((eq n :destroy)
                        (setf (capi:titled-object-message interface) +null-string+))
                       (t (unless n
                            (setq n (if (numberp (capi:interface-message-area interface))
                                        (1+ (capi:interface-message-area interface))
                                        0))
                            (unless title
                              (when-let* ((string (capi:titled-object-message interface))
                                          (start (char-position #\Rubout string)))
                                (when (< 1 start)
                                  (setq title (subseq string 0 (1- start)))))))
                          (setf (capi:titled-object-message interface)
                                (lw:string-append (or title +null-string+) " "
                                                  (make-string n :initial-element #\Rubout
                                                               :element-type 'base-char))
                                (capi:interface-message-area interface) n))))
               interface n title))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PROGRESS-DIALOG-BASE  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Initial release by Michal Krupka, October 2004 (MichalKrupka<at>mac<dot>com)
;;; Differences from progress-meter
;;; - Has the Cancel button.
;;; - The run function is executed in dedicated process, called "Progress runner xxx".
;;;
;;; Differences from Michal's
;;; - Exported progress-cancel function.
;;; - Added cancel-text slot, which stores the original cancel-button text
;;;   (used in the wizard-interface subclass).
;;; - Added initarg :cancel-p applied to the cancel-button.
;;; - Added initarg :cancel-function that specifies a selection-callback
;;;   for the cancel-button (progress-cancel is the default).
;;; - Added the following initargs to display-progress-dialog:
;;;   -- :modal,
;;;   -- :owner, which passed the capi:display-dialog when modal is true
;;;      (but is not passed to the run function).
;;; - Protected (apply %function %args) with ywi:top-level-hook.
;;; - Augmented the abort restart report function.
;;;
;;; Overview
;;; --------
;;; The function display-progress-dialog displays a simple dialog with a message,
;;; cancel button, and optional progress-bar. It accepts a run-function as its
;;; first parameter, which will be run in a separate process while the dialog
;;; is displayed.
;;;
;;; When the programmer-supplied run-function returns, the dialog dismisses
;;; and, if modal, returns the result of the function as its first value.
;;;
;;; There is also a possibility to update the message and attributes of
;;; the progress bar in the dialog using the function progress-change.
;;;
;;; Execution abort
;;; - Invoking the abort function from inside the run-runction results in
;;;   termination immediately.
;;; - Pressing the Cancel button or invoking the progress-cancel function
;;;   defers termination until the next call of progress-change.
;;; The run-function can also return :abort to inform GUI and caller that
;;; its execution was abnormal.
;;;
;;; NB: We do not terminate the runner process by means of mp:process-kill.

;; Variable dynamically bound to the progress dialog instance during the execution
;; of the run-function:
;; - supplies a default for progress-change function,
;; - allows user communication through dialogs owned by the progress dialog.
(defvar *progress-dialog* nil)

(capi:define-interface progress-dialog-base ()
 ((progress-bar :reader progress-bar :initform nil)
  (cancel-text :initform #L"Abort")
  (cancel-message :initform #L"Stopping..." :initarg :cancel-message)
  (process :initarg :dedicated-process :initform t)
  (%mailbox% :initform (mp:make-mailbox)))
 (:panes
  (message-pane capi:message-pane
	:visible-min-width nil :visible-max-width nil
	:visible-min-height '(:character 2))
  (cancel-button capi:push-button :mnemonic-text #L"Abort" 	; don't auto underline
	:callback 'progress-cancel :callback-type :interface))
 (:default-initargs
  :confirm-destroy-function 'progress-cancel))

(defmethod initialize-instance :after ((self progress-dialog-base)
                                       &key (progress t)
                                            ((:cancel-button text) t supplied-p)
                                            cancel-p
                                            (cancel-function 'progress-cancel))
  (with-slots (progress-bar cancel-text cancel-button) self
    (when progress
      (setf progress-bar (apply #'make-instance 'capi:progress-bar
                                (if (consp progress) progress ()))))
    (setf (capi:button-cancel-p cancel-button) cancel-p
          (capi:callbacks-selection-callback cancel-button) cancel-function)
    (cond ((null text)
           (setf (capi:button-enabled cancel-button) nil))
          ((stringp text)
           (allf text (capi:item-text cancel-button) cancel-text))
          (supplied-p
           (setf (capi:item-text cancel-button) cancel-text)))))	; default text


;; Generics
;; - Major executer of a run-function, which can call progress-change periodically.
;; - Result code analyzer that can exit the dialog or modify its view some way.
;;   Args: result Either the first value returned by the progress-call,
;;                or :abort = abnormal termination.
(defgeneric progress-call (self &rest args))
(defgeneric progress-complete (self result))

(defun progress-run (self &rest args)
 ;;; Start runner process as soon as the dialog is ready for display,
  ;; Args: args  Passed to progress-call
  ;; Usage: E.g. as create-callback.
  ;; NB: The current process of the dialog will remain mp:mailbox-reader-process.
  (with-slots (%mailbox% process) self
    (let ((mp:*process-initial-bindings* (acons '*progress-dialog* self
                                                mp:*process-initial-bindings*)))
      (setf process (mp:process-run-function
                       (format nil "Progress runner ~d" (get-internal-run-time))
                       (list :mailbox %mailbox%)
                       '%pdr-startup self args)))))

(defun progress-alive-p (self)
  (with-slots (process) self
    (and (mp:process-p process) (mp:process-alive-p process))))

(defun %progress-set-status (&key (interface *progress-dialog*)
                                  message start end position delta)
  (capi:execute-with-interface interface
    (lambda ()
      (when message
        (setf (capi:title-pane-text (slot-value interface 'message-pane)) message))
      (when-let (progress-bar (progress-bar interface))
        (when start     (setf (capi:range-start progress-bar) start))
        (when end       (setf (capi:range-end progress-bar) end))
        (cond (position (setf (capi:range-slug-start progress-bar)
                               (if (eq position :end)
                                   (capi:range-end progress-bar)
                                   position)))
              (delta    (incf (capi:range-slug-start progress-bar) delta)))))))

(defun progress-cancel (self &optional (disable t))
 ;;; This function must is called in the dialog process and lets interactive
  ;; abort be handled by the runner process (later on next %pdr-maybe-abort call).
  ;; Used as: selection-callback of cancel-button.
  (with-slots (%mailbox% cancel-button cancel-message) self
    (%progress-set-status :interface self :message cancel-message)
    (when disable (disable cancel-button))
    (mp:mailbox-send %mailbox% :abort))
  nil)

;;; Helpers executed in "Progress runner" process

(defun %pdr-maybe-abort ()
  (when (eql (mp:mailbox-peek (mp:process-mailbox mp:*current-process*)) :abort)
    (abort)))

(defun %pdr-startup (self args)
 ;;; Only the final test is being run in the progress-dialog interface proccess.
  ;; The runner function is invoked in the "Progress runner" (current process).
  ;; Args: self  The progress dialog or interface that may have not be displayed yet.
  ;;       args  Passed to progress-call
  (capi:execute-with-interface self
   #'progress-complete
   self
   (restart-case (with-restarts (:quit nil)		; prohibit quiting pdr abnormally
                   (prog1 ;(cons :exit
                       (top-level-hook (lambda () (apply #'progress-call self args)) self)
                     (%pdr-maybe-abort)))
     (abort () :report (list "~a ~a" #L"Abort" (capi:interface-title self))
            :abort)))
  (setf (slot-value self 'process) nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  PROGRESS-DIALOG  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(capi:define-interface progress-dialog (progress-dialog-base)
 ((%function :initarg :function)
  (%args :initarg :args :initform nil))
 (:layouts
  (main-layout capi:column-layout '(message-pane buttons) :internal-border 4)
  (buttons capi:row-layout '(nil cancel-button)))
 (:default-initargs
  :layout 'main-layout
  :auto-menus nil
  :visible-min-width '(:character 40)
  :create-callback 'progress-run))

(defmethod initialize-instance :after ((self progress-dialog)
                                       &key message start end position)
  (with-slots (main-layout progress-bar) self
    (when progress-bar
      (setf (capi:layout-description main-layout)
            (insert 'progress-bar (capi:layout-description main-layout) :after 0))))
  (%progress-set-status :interface self
                        :message message :start start :end end :position position))

#|(defmethod interface-confirm-destroy ((self progress-dialog) &optional cancelable
                                      &rest args)
  (declare (ignore cancelable args))
  (progress-cancel self))	; => NIL, i.e. do not destroy |#

(defmethod progress-call ((self progress-dialog) &rest args)
  (declare (ignore args))
  (apply (slot-value self '%function) (slot-value self '%args)))

(defmethod progress-complete ((self progress-dialog) result)
  (if (eq result :abort)
      (capi:abort-dialog)
      (capi:exit-dialog result))
  (capi:destroy self))			; we get here only if the dialog is non-modal
  ;(ecase (first-or-self result)
  ;  (:exit  (capi:exit-dialog (if (consp result) (cdr result) nil)))
  ;  (:abort (capi:abort-dialog)))

;;; Top-level API

(defun progress-change (&rest args)
 ;;; This can be called inside the run function to update the progress dialog panes.
  ;; Key args:
  ;;   message	NIL or string displayed as the message in the dialog.
  ;;   start	NIL or number specifying a new minimal value of the progress bar.
  ;;   end	NIL or number specifying a new maximal value.
  ;;   position	NIL or number specifying a new actual value.
  ;;   delta	NIL or number specifying an increase to the slug position.
  ;; When none of the arguments is supplied, :delta 1 is assumed by default. 
  ;;
  ;; This function first checks wheather the Cancel button has been clicked.
  ;; If so, it invokes the installed abort restart and close the dialog. 
  ;; If the Cancel button has not been clicked, the dialog panes are updated in according
  ;; to the args supplied.
  ;;
  ;; Called by: Run function of any progress-dialog-base subclass, e.g. wizard-interface.
  (%pdr-maybe-abort)
  (apply #'%progress-set-status (or args '(:delta 1))))

(defun display-progress-dialog (function args &rest interface-args
                                              &key (modal t) owner
                                                   (best-width '(/ :screen-width 2))
                                                   (best-x '(/ :screen-width 4))
                                                   (best-y '(/ :screen-height 3))
                                              &allow-other-keys)
 ;;; Executes function in a dedicated process and displays a progress-indicating dialog.
  ;; Args:
  ;;  function		A function to run.
  ;;  args		List of arguments passed to run-function.
  ;;  interface-args	A list of arguments to be used when creating an instance of an
  ;;			internal interface class; besides the following arguments,
  ;;			any argument acceptable by capi:interface can be used.
  ;;  message		A string containing a message to be displayed in the dialog 
  ;;                    (default "").
  ;;  progress-bar	If True, the dialog will also contain a progress bar (default nil).
  ;;  start		Minimal value of the progress bar (default 0).
  ;;  end		Maximal value of the progress bar (default 100).
  ;;  position		Initial value of the progress bar (default 0).
  ;;  cancel-button	Text to appear on the cancel button (default "Cancel").
  ;;  cancel-message	Message to appear in the dialog when user clicks the cancel button
  ;;                    (default "Stopping...").
  ;; Values:
  ;;  1) result		Primary value of run-function or nil.
  ;;  2) successp	T if run-function returned normally,
  ;;			NIL if abort restart has been chosen (either directly in 
  ;;			the run-function or by clicking the cancel button).
  ;;  or  The dialog interface object itself is returned imediately if modal is false.
  ;; NB:
  ;;  - run-function should call the function progress-change periodically
  ;;	to update the message and attributes of the progress bar in the dialog
  ;;	and to check wheather the cancel button has been clicked.
  ;;  -	During the execution of run-function, an abort restart is installed.
  ;;	When it is invoked, the function display-progress-dialog closes
  ;;	the progress dialog and returns values nil, nil.
  (let ((self (apply #'make-instance 'progress-dialog
                     :function function  :args args
                     :best-width best-width  :best-x best-x  :best-y best-y
                     (sans interface-args :modal :owner :best-width :best-x :best-y))))
    (if modal
        (capi:display-dialog self :owner owner)
        (capi:display self))))
