;;;-*- Mode: Lisp; -*-
;;; Ystok-Widgets - LispWorks CAPI-based GUI framework
;;; Copyright (c) 2003-2010 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Application framework generics, toolbars, and menus
;;; Toolbar NB: In button tooltips, mnemonic chars are ignored automatically.

(in-package :ystok.widgets)

;;; Helpers

(defun beep-not-implemented (pane &optional operation)
  (capi:beep-pane (capi:top-level-interface pane))
  (message-area-format pane #L"Operation~@[ '~a'~] is not available in this version."
                       operation))

(defun beep-impossible (pane &optional operation &key (name nil specified-p) condition)
 ;;; Args: pane      May be NIL
  ;;	   operation Symbol, stirng, or NIL.
  ;;       name      Explicit name, title, or arbitrary object that we has failed
  ;;		     to operated on (formated with ~A).
  ;;		     If not specified, pane title is tried instead.
  ;;	   condition Explanation as an instance of condition or string.
  (capi:beep-pane (if pane (capi:top-level-interface pane)))
  (when pane
    (message-area-format pane #L"Operation~@[ '~a'~] is impossible~@[ for ~s~]~@[: ~a~]."
                       (if operation
                           (remove #\& (string
                                        (case operation
                                          (:cut        #L"Cu&t")
                                          (:copy       #L"&Copy")
                                          (:paste      #L"&Paste")
                                          (:select-all #L"Select &All")
                                          (:undo       #L"&Undo")
                                          (otherwise operation))))
                           nil)
                       (if specified-p
                           name
                           (or (object-title pane) (capi:capi-object-name pane)
                               (type-of pane)))
                       (typecase condition
                         (condition (princ-to-string condition))
                         (string condition)))))

;;; Interface generic functions
;;; NB: Callbacks create-callback, confirm-destroy-function, and destroy-callback
;;;     are called automatically only for top-level interfaces.

(defgeneric interface-on-create (interface)
 (:method (interface)))

(defgeneric interface-confirm-destroy (interface &optional cancelable &rest args)
 ;;; Confirm action of
 ;;; - either closing the window at all,
  ;; - or unloading it (i.e. clearing from any object contained),
  ;; - or hiding it (e.g. called by cancel-choice or take-choice).
  ;; Args: cancelable  If true, the action is cancellable, e.g. cancel button is enabled.
  ;;       args        More key args passed to save-interface (and revert-interface?)
  ;; Default method get us the freedom of invoking (call-next-method).
 (:method (interface &optional (cancelable t) &rest args)
  (declare (ignore interface cancelable args))
  t))

(defgeneric interface-on-destroy (interface)
 (:method-combination progn)
 (:method progn (interface)))

(defgeneric operate-on-interface (interface operation &optional focus-pane)
 (:method (interface operation &optional focus-pane)
  (beep-not-implemented interface operation)))

(defgeneric save-interface (interface &optional new-focus &key &allow-other-keys)
 ;;; Save all panes and grid records into corresponding underlaying object(s).
  ;; Args: interface Instance of ywi:data-interface and/or tracking-interface-mixin.
  ;;	   focus     New document or pane we going to move focus to.
  ;; Value: True = success, False = failure.
  ;; Used by: :callback 'save-interface  :callback-type :interface-data/(:interface :focus)
  ;; Q: Is realm parameter passed anywhere? Do we need :callback-type :interface-realm?
 (:method (interface &optional focus &key)
  (declare (ignore focus))
  (beep-not-implemented interface :save)))

;;; Pane generic and convenience functions

(defgeneric operate-on-focus (focus-pane operation)
 ;;; Args: operation Except "standard" edit/select family, i.e. neither
  ;;                 :cut, :copy, :paste, :undo, :select-all, nor :deselect-all
 (:method (pane operation)
  (beep-impossible pane operation)))

(defun show (&rest panes)
  (dolist (pane panes) (setf (shown pane) t)))
(defun hide (&rest panes)
  (dolist (pane panes) (setf (shown pane) nil)))

(defun enable (&rest panes)
  (dolist (pane panes) (setf (enabled pane) t)))
(defun disable (&rest panes)
  (dolist (pane panes) (setf (enabled pane) nil)))

(defgeneric update-toolbars (parent &rest names)
 ;;; Args: names List of symbols or capi:toolbar instances;
  ;;             can contain NILs (for featured forms)
 (:method (self &rest names)
  (declare (ignore self names)))
 (:method ((self capi:toolbar) &rest names)
  (capi:update-toolbar self)
  (mapc #'capi:update-toolbar names))			; all must be toolbar instances
 (:method ((parent capi:layout) &rest names)
  (dolist (name names)
    (when-let (toolbar (if (symbolp name) (find-pane-by-name parent name) name))
      (capi:update-toolbar toolbar))))
 (:method ((parent capi:docking-layout) &rest names)
  (let ((items (capi:docking-layout-items parent)))	; find-pane-by-name misses
    (dolist (name names)				; undocked toolbars
      (when name
        (when-let (toolbar (if (symbolp name) 
                               (car (assoq name items :key #'capi:capi-object-name))
                               name))
          (capi:update-toolbar toolbar))))))
 (:method ((interface capi:interface) &rest names)
  ;; Args: names List of slot names or toolbar instances
  (dolist (name names)
    (when name
      (when-let (toolbar (if (symbolp name) (slot-value interface name) name))
        (capi:update-toolbar toolbar))))) )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  EDIT-TOOLBAR  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; If we have defined accelerators like Ctrl+X, they would have stopped working
;;; for instances of text-input-pane and subclasses as there is no way to pass
;;; the corresponding events there.
;;; NB: Updating the toolbar in "real-time" after every edit gesture seems to be a pain.
;;;	So :enabled-function is specified for the edit-menu but is omitted for the toolbar.

(defun edit-image-set ()
  (capi:make-general-image-set :id #.(gp:read-external-image
                                      (lw:current-pathname "images/edit.bmp")
                                      :transparent-color-index 13)
                               :image-count 6 :width 16 :height 16))
;; CAUTION: Symbol macros does not work in defclass :default-initargs!
(define-symbol-macro *edit-image-set* (edit-image-set))
 #|(capi:make-general-image-set :id #.(gp:read-external-image
                                      (lw:current-pathname "images/edit.bmp")
                                      :transparent-color-index 13)
                               :image-count 6  :width 16  :height 16))|#

(defgeneric pane-complete-p (pane)
 ;;; Predicate returning true if completion is possible 
 (:method (pane) nil)
 (:method ((pane capi:editor-pane)) t)
 (:method ((pane capi:text-input-pane))
  (capi:text-input-pane-completion-function pane)) )

(defun active-pane-complete-p (&optional pane)
  (pane-complete-p (get-active-pane-interface pane)))
(defun active-pane-complete (&optional pane)
  (let ((pane (get-active-pane-interface pane)))
    (if (pane-complete-p pane)
        (pane-complete pane)
        (beep (format nil #L"Completion function is undefined for pane or column~@[ ~s~]."
                      (object-title pane))
              pane))))


(defgeneric edit-toolbar (interface)
 ;;; Get/find the Edit toolbar of the interface
 (:method (interface) nil))

(defun make-edit-toolbar (&key (title #L"Edit") (callback 'operate-on-focus)
                               clear complete
                               interface active)
 ;;; Create the Edit toolbar for the instance.
  ;; Args: clear     If true, provide Clear button
  ;;       complete  If true, provide Complete button
  ;;	   interface If specified, fill its edit-toolbar slot with the toolbar created
  ;;	   active    If true, enabled function is set for most of the buttons and
  ;;                 this assumes that capi:update-toolbar should be called frequently.
  ;; Value: Tree-element list that are pushed into docking-layout description.
  (let ((toolbar
         (make-instance 'capi:toolbar :name 'edit-toolbar
          :items
          `(,(make-instance 'capi:toolbar-component	;:name 'edit-standard-tc
              :items `(,(make-instance 'capi:toolbar-button :data :cut
			:image 0
			:callback-type #1='(:focus)
                        :selection-callback 'capi:active-pane-cut
                        :enabled-function (if active 'capi:active-pane-cut-p nil)
			:tooltip #-lispworks4.3 #L"Cu&t" #+lispworks4.3 #L"Cut")
                       ,(make-instance 'capi:toolbar-button :data :copy
			:image 1
			:callback-type #1#
                        :callback 'capi:active-pane-copy
			:enabled-function (if active 'capi:active-pane-copy-p nil)
			:tooltip #-lispworks4.3 #L"&Copy" #+lispworks4.3 #L"Copy")
                       ,(make-instance 'capi:toolbar-button ;:remapped :paste
			:data :paste  :image 2
			:callback-type #1#
                        :callback 'capi:active-pane-paste
			:enabled-function (if active 'capi:active-pane-paste-p nil)
			:tooltip #-lispworks4.3 #L"&Paste" #+lispworks4.3 #L"Paste")))
            ,(make-instance 'capi:toolbar-button :data :undo
		:image 3
		:callback-type #1#
		:callback 'capi:active-pane-undo
		:enabled-function (if active 'capi:active-pane-undo-p nil)
		:tooltip #-lispworks4.3 #L"&Undo" #+lispworks4.3 #L"Undo")
            ,@(when (or clear complete)
              `(,(make-instance 'capi:toolbar-component
		:items
                `(,@(when clear
                    `(,(make-instance 'capi:toolbar-button :data :clear
			:image 5
			:callback-type '(:focus :data)
			:callback 'operate-on-focus
			:tooltip #L"E&rase")))
                  ,@(when complete
                    `(,(make-instance 'capi:toolbar-button
			:image 4
			:callback-type #1#
			:callback 'active-pane-complete
			:enabled-function (if active 'active-pane-complete-p nil)
			:tooltip #L"Comp&lete"))))))))
          :selection-callback callback
	  :callback-type '(:focus :data)
          :default-image-set (edit-image-set)
          :button-width 18 :button-height 18)))
    (when (and interface (slot-exists-p interface 'edit-toolbar))
      (setf (slot-value interface 'edit-toolbar) toolbar))
    (list toolbar :title title)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  EDIT-MENU  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; To preserve the "edit/select" functionality built in CAPI 4.x/5.x.,
;;; we do not assign "real" accelerators and mimic them in titles.

(defun edit-menu (interface &key (callback 'operate-on-focus)
                                 (select-all #\Shift-Control-\a)
                                 #+ys-product zoom)
  (make-instance 'capi:menu
   :title #L"Edit"  :mnemonic 0
   :name 'edit-menu
   :items (edit-menu-items :select-all select-all #+ys-product #+ys-product :zoom zoom)
   :callback-type '(:focus :data)
   :selection-callback callback
   :menu-interface interface
   :help-key nil))

#-lispworks4.3
(defun edit-menu-items (&key popup (select-all #\Shift-Control-\a)
                             #+ys-product zoom)
 ;;; Args: popup True if used for context menu
  ;;       zoom  Callback for zoom-mi
 `(,(make-instance 'capi:menu-component :items
    `(,(make-instance 'capi:menu-item
	:mnemonic-title (if popup
                            #L"&Undo"
                           (load-time-value (string-append #L"&Undo" #\Tab
                                             (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                              (sys:coerce-to-gesture-spec #\Control-\z)))))
	:data :undo  :name :undo ;:accelerator #\control-\z
	:callback-type #1='(:focus)
	:selection-callback 'capi:active-pane-undo
	:enabled-function 'capi:active-pane-undo-p
	:help-key nil)
      ,@(unless popup
          `(,(make-instance 'capi:menu-item :mnemonic-title #L"Comp&lete"
		:name :complete
		:accelerator *completion-gesture*
		:callback-type #1#
		:selection-callback 'active-pane-complete
		:enabled-function 'active-pane-complete-p
		:help-key *completion-help-key*)))))
   ,(make-instance 'capi:menu-component :items 	;:name edit-standard-mc
    `(,(make-instance 'capi:menu-item
	:mnemonic-title (if popup
                            #L"Cu&t"
                          (load-time-value (string-append #L"Cu&t" #\Tab
                                            (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                             (sys:coerce-to-gesture-spec #\Control-\x)))))
	:data :cut  :name :cut
	:callback-type #1#
	:selection-callback 'capi:active-pane-cut
	:enabled-function 'capi:active-pane-cut-p
	:help-key nil)
      ,(make-instance 'capi:menu-item
	:mnemonic-title (if popup
                            #L"&Copy"
                          (load-time-value (string-append #L"&Copy" #\Tab
                                            (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                             (sys:coerce-to-gesture-spec #\Control-\c)))))
	:data :copy  :name :copy
	:callback-type #1#
	:selection-callback 'capi:active-pane-copy
	:enabled-function 'capi:active-pane-copy-p
	:help-key nil)
      ,(make-instance 'capi:menu-item
	:mnemonic-title (if popup
                            #L"&Paste"			; -> sys:print-pretty-gesture-spec
                          (load-time-value (string-append #L"&Paste" #\Tab
                                            (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                             (sys:coerce-to-gesture-spec #\Control-\v)))))
	:data :paste :name :paste
	:callback-type #1#
	:selection-callback 'capi:active-pane-paste
	:enabled-function 'capi:active-pane-paste-p
	:help-key nil) ))
   ,@(when select-all
       `(,(make-instance 'capi:menu-component :items 	;:name ;edit-select-mc
	   `(,(make-instance 'capi:menu-item
		:mnemonic-title #L"Select &All"
		:data :select-all
		:accelerator (if popup nil select-all)
		:callback-type #1#
		:selection-callback 'capi:active-pane-select-all
                :enabled-function 'capi:active-pane-select-all-p
		:help-key nil)
             #+ys-product
             ,@(when zoom
                 `(,(apply #'zoom-mi :popup popup
                           (unless (eq zoom t)
                             `(,:callback ,zoom))))) )))) ))

;;; LW 4.2,4.3: Neither :mnemonic :default nor :mnemonic-title works.
#+lispworks4.3
(defun edit-menu-items (&key popup (select-all #\Shift-Control-\a)
                             #+ys-product zoom)
 `(,(make-instance 'capi:menu-component :items
    `(,(make-instance 'capi:menu-item
	:title (if popup
                   #L"Undo"
                 (load-time-value (string-append #L"Undo" #\Tab
                                             (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                              (sys:coerce-to-gesture-spec #\Control-\z)))))
	:mnemonic 0        
	:data :undo  :name :undo ;:accelerator #\control-\z
	:callback-type #1='(:focus)
	:selection-callback 'capi:active-pane-undo
	:enabled-function 'capi:active-pane-undo-p
	:help-key nil)
      ,@(unless popup
          `(,(make-instance 'capi:menu-item
		:title #L"Complete"
		:name :complete  :accelerator *completion-gesture*
		:callback-type #1#
		:selection-callback 'active-pane-complete
		:enabled-function 'active-pane-complete-p
		:help-key *completion-help-key*)))))
   ,(make-instance 'capi:menu-component :items 	;:name edit-standard-mc
    `(,(make-instance 'capi:menu-item
	:title (if popup
                   #L"Cut"
                 (load-time-value (string-append #L"Cut" #\Tab
                                            (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                             (sys:coerce-to-gesture-spec #\Control-\x)))))
	:data :cut  :name :cut
	:callback-type #1#
	:selection-callback 'capi:active-pane-cut
	:enabled-function 'capi:active-pane-cut-p
	:help-key nil)
      ,(make-instance 'capi:menu-item
	:title (if popup
                   #L"Copy"
                 (load-time-value (string-append #L"Copy" #\Tab
                                            (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                             (sys:coerce-to-gesture-spec #\Control-\c)))))
	:data :copy  :name :copy
	:callback-type #1#
	:selection-callback 'capi:active-pane-copy
	:enabled-function 'capi:active-pane-copy-p
	:help-key nil)
      ,(make-instance 'capi:menu-item
	:title (if popup
                   #L"Paste"
                 (load-time-value (string-append #L"Paste" #\Tab
                                            (SYSTEM::GESTURE-SPEC-AS-MENU-ACCELERATOR
                                             (sys:coerce-to-gesture-spec #\Control-\v)))))
	:data :paste :name :paste
	:callback-type #1#
	:selection-callback 'capi:active-pane-paste
	:enabled-function 'capi:active-pane-paste-p
	:help-key nil) ))
   ,@(when select-all
       `(,(make-instance 'capi:menu-component :items 	;:name ;edit-select-mc
	   `(,(make-instance 'capi:menu-item
		:title #L"Select all"
		:data :select-all  :accelerator select-all
		:callback-type #1#
		:selection-callback 'capi:active-pane-select-all
                :enabled-function 'capi:active-pane-select-all-p
		:help-key nil)
             #+ys-product
             ,@(when zoom
                 `(,(zoom-mi :callback zoom))) )))) ))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DOCUMENT (FILE or OBJECT)  ;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Components and items for toolbars named File, Document, or the like

(defun doc-image-set (&optional index &key (image-count 13))
  (let ((image-set (capi:make-general-image-set
                    :id #.(gp:read-external-image
                           (lw:current-pathname "images/doc.bmp")
                           :transparent-color-index 13)
                    :image-count image-count  :width 16  :height 16)))
    (if index
        (capi:make-image-locator :image-set image-set :index index)
        image-set)))

(define-symbol-macro *doc-image-set* (doc-image-set))
#|(capi:make-general-image-set :id #.(gp:read-external-image	; LispWorks 6.0 only
                                      (lw:current-pathname "images/doc.bmp")
                                      :transparent-color-index 13)
                               :image-count 13  :width 16  :height 16))|#

(defun doc-print-tc (&key (callback 'operate-on-interface)
                          (callback-type '(:interface :data :focus))
                          (default-image-set *doc-image-set*) (images '(4 5)))
  (make-instance 'capi:toolbar-component
   :items `(:print :preview)  :tooltips '(#L"Print" #L"Preview")  :images images
   :default-image-set default-image-set
   :selection-callback callback
   :callback-type callback-type))

(defun doc-print-mc (interface &key (callback 'operate-on-interface)
                                    (callback-type '(:interface :data :focus))
                                    (name 'doc-print-mc)
                                    (accelerator #\Control-\p)
                                    document
                                    data enabled-function ellipsis)
 ;;; Args: data     Odd list of other arguments with last element eq to :operation
  ;;	   ellipsis String or true means append the suffix to standard titles
  (declare (ignore interface))
  (unless document (setq document #L"document"))
  (make-instance 'capi:menu-component :items
    `(,(make-instance 'capi:menu-item 
                      :title (if ellipsis
                                 (string-append #L"Print" #1=(if (stringp ellipsis)
                                                                 ellipsis "..."))
                                 #L"Print")
                      :mnemonic 0
                      :data (if data (append (listify data) (list :print)) :print)
                      :accelerator accelerator
                      :help-key (format nil #L"Output ~a directly to default printer."
                                        document))
      ,(make-instance 'capi:menu-item 
                      :title (if ellipsis (string-append #L"Preview" #1#) #L"Preview")
                      :mnemonic 4
                      :data (if data (append (listify data) (list :preview)) :preview)
                      :help-key (format nil #L"Show ~a on screen in printer-ready form."
                                        document)))
    :name name
    :callback callback  :callback-type callback-type
    :enabled-function enabled-function))

(defun doc-refresh-tc (&key (callback 'operate-on-interface) (help-key #L"Refresh"))
                            ;(images '(9))
  (make-instance 'capi:toolbar-component
   :items `(:refresh)
   :tooltips `(,help-key)
   :images '(0) ;images
   :default-image-set (capi:make-general-image-set
                       :id #.(gp:read-external-image
                              (lw:current-pathname "images/refresh.bmp")
                              :transparent-color-index 13)
                       :image-count 1 :width 16  :height 16)
   :selection-callback callback  :callback-type '(:interface :data :focus)))

(defun doc-refresh-mi (&key (callback 'operate-on-interface) enabled-function
      (help-key #L"Update window or recalculate document, maybe by requering database."))
  (make-instance 'capi:menu-item  :title #L"Refresh" :mnemonic 0  :accelerator #\F5
    :data :refresh
    :enabled-function enabled-function
    :callback callback  :callback-type '(:interface :data :focus)
    :help-key help-key))

;;; Editor shortcuts used instead of/along with accelerators

(editor:defcommand "New" (p)
 "Create new document or object in the current top-level interface" ""
  (declare (ignore p))
  (let ((pane (editor::current-editor-pane)))
    (operate-on-interface (capi:top-level-interface pane) :new pane)))

(editor:defcommand "Open" (p)
 "Open document or object in the current top-level interface" ""
  (declare (ignore p))
  (let ((pane (editor::current-editor-pane)))
    (operate-on-interface (capi:top-level-interface pane) :open pane)))

(editor:defcommand "Print" (p)
 "Print document or object displayed in the current top-level interface" ""
  ;; CAUTION: For rr:menu-run callback, callback-type is :full.
  ;;          Can grid be embedded into report-board?
  (declare (ignore p))
  (let ((pane (editor::current-editor-pane)))
    (operate-on-interface (capi:top-level-interface pane) :print pane)))

(editor:defcommand "Save" (p)
 "Save document or object displayed in the current top-level interface" ""
  (declare (ignore p))
  (save-interface (capi:top-level-interface (editor::current-editor-pane))))
