;;;-*- Mode: Lisp; -*-
;;; Ystok-Widgets - LispWorks CAPI-based GUI framework
;;; Copyright (c) 2003-2010 Dr. Dmitriy Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Constants, variables, generic and core functions

(in-package :ystok.widgets)

(defconstant +none-circle+ '#1=(:none . #1#))	; cicled list of :none for :mnemonics

(defconstant +all-file-filters+ '(#L"All files" "*.*"))
(defconstant +all-none-list+ (list +all-string+ +none-string+))
(defconstant +items-uninitialized+ #())	; Special items value for a drop-down list
					; to be filled in dynamically
;;; Pixel sizes. GUIDELINES: 1 DLU = 1/4 font width = 1/8 font height
;;; Q: Shall we turn them into variables and initialize on starting up the application?
(defconstant +h-gap+ 6)			; 4DLUs between ordinary panes in row-layout
(defconstant +big-h-gap+ 11)		; 7DLUs between unrelated/boxes in row-layout
(defconstant +v-gap+ 6)			; 4DLUs between weakly related in column-layout
(defconstant +big-v-gap+ 11)		; 7DLUs between unrelated/boxes in column-layout
(defconstant +padding+ 6)		; 7DLUs on sides of dialog box or tab page

(defvar *erroneous-pane* nil)	; pane to blame for last error signaled

(defvar *message-area* t)	; show message area in top-level interfaces
(defvar *show-help-button* t)	; show Help button in modal dialogs even if '?' contexthelp
(defvar *show-tooltips* t)	; has no effect to toolbars, but only to client area panes

(defgeneric slot-name (object)
 (:method (object) nil)
 (:method ((pane capi:simple-pane))
  (or (capi:capi-object-property pane :slot-name)
      (capi:capi-object-name pane)))
 (:method ((pane capi:toolbar-object))
  (or (capi:capi-object-property pane :slot-name)
      (capi:capi-object-name pane))) )

;;; Generic completion feature
;;; *completion-gesture* should be a menu item accelerator
;;; pane-complete must be invoked in the pane process (is pane obligatory focused?)

(defparameter *completion-gesture* #\Control-\Space)
(defparameter *completion-help-key* #L"Complete text automatically or pop up a choice window.")

(defgeneric pane-complete (pane)
 (:method (pane)
  (message-area-format pane
    #L"Completion function is undefined for pane or column~@[ ~s~]."
    (object-title pane))
  (beep))		; E.g. text-input-choice, which cannot be in focus
 (:method ((pane capi:editor-pane))
  (capi:call-editor pane *completion-gesture*)) ; "Complete Symbol"
 (:method ((pane capi:text-input-pane))
  (capi:text-input-pane-complete-text pane)) )

(defun current-top-level-interface ()
 ;;; Value: Active application window or NIL if another application is active.
 ;(capi::default-interface)?
 #+win32 (when-let (window ww::*active-window*) (win32::r-top-element window))
 #-win32 (first (capi:collect-interfaces 'capi:interface :screen :any :sort-by :visible)))

;;; CAUTION:
;;;  Changing capi:simple-pane-enabled for layout can lead to focus misbehavior,
;;;  e.g. cause for the top-level interface not to receive focus via a mouse click.

(defgeneric enabled (object)
 (:method (object)  (declare (ignore object))  t)
 (:method ((face capi:interface))
  (if (capi:top-level-interface-p face)
      (getf (capi-internals:capi-object-plist face) 'enabled t)
      (call-next-method)))
 (:method ((pane capi:simple-pane))	 (capi:simple-pane-enabled pane)) 
 (:method ((object capi:toolbar-object)) (capi:simple-pane-enabled object))
 (:method ((object capi:menu-object))	 (capi:menu-object-enabled object)) )

(defgeneric (setf enabled) (value object)
 (:method (value object)		; needed for map-pane-children not to signal
  (declare (ignore object))		; capi:capi-object preceeds capi:toolbar-object!
  value)

 (:method (value (face capi:interface))
  (if (capi:top-level-interface-p face)
      (setf (capi:capi-object-property face 'enabled) value)
      (call-next-method)))

 (:method (value (pane capi:simple-pane))
  (setf (capi:simple-pane-enabled pane) value))

 (:method :after (value (pane capi:simple-pane))
  (when-let (title-pane (object-title-pane pane))
    (setf (capi:simple-pane-enabled title-pane) value)))

 (:method ((value (eql :read-only)) (object capi:text-input-range))
  (setf (capi:simple-pane-enabled object) nil))		; does not support :read-only

 (:method :after (value (pane capi:text-input-pane))
  ;; When a pane equipped by buttons, affect all the toolbar buttons as well
  (when-let (parent (capi:element-parent pane))
    (when (eq (capi:capi-object-name parent) 'CAPI::TEXT-INPUT-PANE-BUTTON-ROW)
      (setf (capi:text-input-pane-buttons-enabled pane) (if value t nil)))))
      ;(capi:map-pane-children parent
      ;  (lambda (child) (unless (eq child pane)
      ;                    (setf (capi:simple-pane-enabled child) value))))

 (:method :before (value (pane CAPI::PANE-WITH-CHILDREN))
  (let ((enabled (enabled pane)))
    (when (or (xor value enabled)
              (xor (eq value :read-only) (eq enabled :read-only)))
      (capi:map-pane-children pane (lambda (child) (setf (enabled child) value))))))

 (:method (value (object capi:toolbar-object))
  (setf (capi:simple-pane-enabled object) value))
 (:method (value (object capi:menu-object))
  (setf (capi:menu-object-enabled object) value)) )

#+lispworks4.4
(defmethod (setf enabled) (value (object capi:option-pane))
  (cond ((eq value :read-only)
         (setf (capi:option-pane-enabled-positions object) nil))
        ((and value (null (capi:option-pane-enabled-positions object)))
         (setf (capi:option-pane-enabled-positions object) :all)))
  (setf (capi:simple-pane-enabled object) value))


;;; Flag meaning "updated" by the user

(defgeneric changed (object)
 (:method ((pane capi:capi-object))
  (capi:capi-object-property pane :changed))
 (:method ((pane capi:editor-pane))
  (editor:buffer-modified (capi:editor-pane-buffer pane))))

(defgeneric (setf changed) (value object)
 (:method (value (pane capi:capi-object))
  (setf (capi:capi-object-property pane :changed) value))
 (:method (value (pane capi:editor-pane))
  (unless value (capi:call-editor pane "Clear Undo"))
  (setf (editor:buffer-modified (capi:editor-pane-buffer pane)) value)))

;;; Flag denoting that data has is somehow put into the widget (and maybe rendered)
;;; *unbound-value* can indicate "semi-loaded", i.e. something undetermened is loaded.

(defun loaded (object)
  (capi:capi-object-property object 'loaded))
(defun (setf loaded) (value object)
  (setf (capi:capi-object-property object 'loaded) value))

;;; Title

(defgeneric object-title (object)
 ;;; Value: String or symbol (i.e. capi-object-name),
  ;;        or NIL (only for instances of capi::titled-pane),
  ;;	    or object itself (for non-capi-object instances).
  ;; NB: titled-object is not a superclass of capi-object;
  ;;     so the latter may precede the former in CPL!
 (:method (self) self)
 (:method :around ((self capi:capi-object))
  (if-bind (title (capi:capi-object-property self :title))
    (if (eq title :parent)
        (object-title (capi:element-parent self))
        (object-title title))
    (call-next-method)))
 (:method ((self capi:capi-object))	(capi:capi-object-name self))
 (:method ((self capi:titled-object))	(or (capi:titled-object-title self)
                                            (call-next-method)))
 (:method ((self capi::titled-pane))	(capi:titled-object-title self))
 (:method ((self capi:title-pane))	(capi:title-pane-text self))
 (:method ((self CAPI::SIMPLE-BUTTON))	(or (CAPI::MNEMONIC-TEXT self)
                                            (capi:item-data self)  ; isn't a titled-pane!
                                            (call-next-method)))
 ;; capi:titled-object-title does not work for interfaces
 ;; capi:interface-title returns nil for confirmers invoked without explicit title
 (:method ((self capi:interface))	(capi:interface-title self))
 #+win32 (:method ((self win32:r-top))	(object-title (win32::r-top-element self))) )

(defgeneric object-title-pane (object)
 ;;; CAUTION: The :title property can only be assigned either to a pane or to a string!
  ;; NB: capi:capi-object may precede capi:titled-object,
  ;;     so we need a method for titled-pane
 (:method ((self capi:capi-object))
  (capi:capi-object (capi:capi-object-property self :title)))
 (:method ((self CAPI::TITLED-PANE))
  (or (CAPI::TITLED-OBJECT-TITLE-PANE self)
      (capi:capi-object (capi:capi-object-property self :title))
      (let ((parent (capi:element-parent self)))	; as TITLED-OBJECT-TITLE-PANE
        (when (typep parent 'CAPI::FRAME-LAYOUT)	; does not work when within a frame
          (CAPI::FRAME-LAYOUT-TITLE-PANE parent)))))
 (:method ((self capi:interface))
  (unless (capi:top-level-interface-p self)
    (CAPI::TITLED-OBJECT-TITLE-PANE self))) )

(defgeneric (setf object-title) (string object &optional parent)
 ;;; NB: Without the explicit parent, this cannot be called from an after method
  ;;     on initialize-instance, but only on create, display, or tab-page-on-create.
 (:method (string (self capi:capi-object) &optional parent)
  (declare (ignore parent))
  (when-let (pane (object-title-pane self))
    (setf (capi:title-pane-text pane) string))))

(defun (setf object-title-gap) (gap object)
 ;;; NB: This cannot be called from an after method on initialize-instance
  ;;     but only on create, display, or tab-page-on-create.
  (when-let (parent (capi:element-parent object))
    (case (capi:capi-object-name parent)
      #1=(CAPI::TITLE-ROW
          (setf (capi:layout-x-gap parent) gap))
      #2=(CAPI::TITLE-COLUMN
          (setf (capi:layout-y-gap parent) gap))
      (CAPI::TEXT-INPUT-PANE-BUTTON-ROW			; title is usually outside the row
          (when-let (parent (capi:element-parent parent))
            (case (capi:capi-object-name parent)
              #1# #2#))))))

(defgeneric (setf pane-text) (text pane)
 ;;; Programmatic change of the text displayed
 (:method ((text string) (pane capi:text-input-pane))
  (setf (capi:text-input-pane-text pane) text)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  RING COMPATIBILITY  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; In LW 6.*, ring-related symbols are exported from HCL, not SYSTEM package.
;;; But they are still visible as internal in SYSTEM too.

(defun ring-collection-pushnew (item collection)
 ;;; Push an item into a ring-based collection ensuring the item to appear first.
  ;; Pattern of use in define-interface:
  ;;  ((ring :initform (sys::make-ring 20 'ring-name) :allocation :class) ...)
  ;;  (:panes (pane capi:text-input-choice  
  ;;	       :items ring
  ;;	       :items-count-function #'sys::ring-length :items-get-function #'sys::ring-ref
  ;;	       :test-function #'string=) ...)
  (let* ((ring (capi:collection-items collection))
         (position (sys::find-in-ring ring item 0
                                      :test (capi:collection-test-function collection))))
    ;; Remove the previous occurence from the ring if not last
    (when (and position (plusp position))
      (sys::rotate-ring ring position)
      (sys::ring-pop ring)
      (sys::rotate-ring ring (- position)))
    ;; Include the new item into the collection widget's items and update the collection
    (unless (eql position 0)
      (sys::ring-push item ring)
      (setf (capi:collection-items collection) ring
            (capi:choice-selection collection) 0))))

(defgeneric (setf shown) (value pane)
 (:method (value pane)
  (if value (capi:show-pane pane) (capi:hide-pane pane)))
 (:method :after (value (pane capi:simple-pane))
  (when-let (title-pane (object-title-pane pane))
    (if value (capi:show-pane title-pane) (capi:hide-pane title-pane))))
 (:method :after (value (pane capi:text-input-pane))
  ;; For a pane equipped by buttons, affect the toolbar as well.
  ;; Since capi:show-pane is declared to works on simple-panes only
  ;; but toolbar-button is not a simple pane, at least disable/enable it.
  (when-let (parent (capi:element-parent pane))
    (when (eq (capi:capi-object-name parent) 'CAPI::TEXT-INPUT-PANE-BUTTON-ROW)
      (capi:map-pane-children parent
        (lambda (child) (unless (eq child pane)
                          (setf (capi:simple-pane-enabled child) value))))))) )

(defun top-level-hook (function interface)
 ;;; Muffle unwanted errors and handle preferences referred as dynamic variables.
  ;; Otherwise prompt for restart with a custom dialog.
  ;; Args: interface Only used for prompting for restarts and supplies a dialog title;
  ;;		     this function itself can be running in a different process.
  ;; Additionally used by:
  ;;	progress-dialog, methods on settings-dialog.
  ;; NB: ,        .
  ;;	 -   unbound-variable   ?
  (handler-bind
        ((editor:editor-error
          (lambda (condition)
            (apply #'message-area-format interface
                   (simple-condition-format-control condition)
                   (simple-condition-format-arguments condition))
            (beep)
            (abort)))
 #-debug (warning
          (lambda (condition)
            (logg :warn condition)
            (muffle-warning condition)))
 #-debug (simple-error
          (lambda (condition)
            (let ((format (simple-condition-format-control condition)))
              (cond ((and (stringp format)
                          (search "Something wrong with the echo area" format
                                  :test #'string=))
                     (beep)
                     (abort))
                    ;; Throwing out (MP::PROCESS-TAG (:KILL)) when not in message-loop
                    ;; Mes-Id: <200505101215.j4ACFIik064335@lwfs1-cam.cam.lispworks.com>
                    ;; Date: Tue, 10 May 2005 13:15:18 +0100 (BST)
                    ;; Subject: Re: quit-interface in XP
                    ;; From: davef@lispworks.com
                    ;;                  
                    ;; This is always a secondary error, usually caused by an error
                    ;; handler aborting. We've fixed a few of them, but more backtraces
                    ;; might help us to fix others.
                    ((equal (simple-condition-format-arguments condition)
                            '((MP::PROCESS-TAG (:KILL))))
                     (abort))
                    (t
                     (logg :error condition)
                     (prompt-for-restart condition :owner interface))))))
               #|(unbound-variable                 ; some libraries seems to overwrite this
                  #'(lambda (condition)
                      (let ((name (cell-error-name condition)))
                        (if (and (symbolp name) (get name 'yl:thys-preference))
                            (use-value (yl:thys-preference name) condition)
                            (prompt-for-restart condition :owner interface)))))|#
 #-debug (error
          (lambda (condition)
            (logg :error condition)
            (prompt-for-restart condition :owner interface))))
    (funcall function)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  CONVERSION text <=> datum  ;;;;;;;;;;;;;;;;;;;;;;;;;

;; Character to be trimmed from the right or both ends
(defparameter *whitespace-bag* '(#\Space #\Tab #\Newline))

(defun nullable-pane-text (pane)
 ;;; Args: pane  An instance of capi:text-input-pane or string (for convenience)
  (let ((text (string-right-trim *whitespace-bag*
                                 (if (stringp pane)
                                     pane
                                     (capi:text-input-pane-text pane)))))
    (if (plusp (length text)) text nil)))

(defun text-read-converter (arg)
  (if arg (princ-to-string arg) +null-string+))

(defun char-write-converter (text)
  (if (plusp (length text)) (char text 0) #\Space))	; should not be invoked on text=""

(defun real-princ-to-string (arg)
  (if (floatp arg)
      (yl:f-format-to-string arg) ;(yl:e-format-to-string arg)	; TODO: g-format-to-string
      (text-read-converter arg)))
(defun real-princ-to-string-signed (arg)
  (cond ((floatp arg) (yl:f-format-to-string arg :atsign t))
        (arg (if (plusp arg)
                 (string-append #\+ (princ-to-string arg))  ; ~@d does not work for ratios
                 (princ-to-string arg)))
        (t +null-string+)))
(defun real-write-converter (text)
  (yl:parse-number (string-left-trim *whitespace-bag* text)))

;;; The idea of outputting float as integer is nice at the first glance.
;;; But it can cause for processor exceptions when writing via a cursor and
;;; the ODBC descriptor with SQL_C_DOUBLE c-type.
#|(defun integer-princ-to-string (arg)
  (if arg
      (princ-to-string (if (integerp arg) arg (round arg)))
      +null-string+))|#
(defun integer-write-converter (text)
  (parse-integer text))

(defun float-princ-to-string-d2 (arg)
  (yl:f-format-to-string arg :d 2))
(defun float-princ-to-string-d2-signed (arg)
  (yl:f-format-to-string arg :d 2 :atsign t))		; print +xxx for non-negative
(defun float-princ-to-string-d3 (arg)
  (yl:f-format-to-string arg :d 3))
(defun float-write-converter (text)
  (float (yl:parse-number (string-left-trim *whitespace-bag* text))))

(defun $-princ-to-string-signed (arg &rest args)
  (apply #'yl:$-format-to-string arg :atsign t args))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  HELP  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The help-key slot defined in the CAPI::SIMPLE-ELEMENT class contains one of
;;;  (topic-id context-id . "tooltip")
;;;	Full specification.
;;;  (id . "tooltip")
;;;	ID supplies both topic-id and context-id
;;;	Denotes the same anchor or similar paths.
;;;  "tooltip"
;;;	Tooltip only
;;;  id
;;;	For interfaces, only id or (topic-id context-id) make sense
;;;	as there are no tooltips
;;;  :parent
;;;	Redirects request to a reasonbale parent of the element.
;;; Q: Could the last context retrieved be memoized
;;;	- either by modifying the (dotted) list desructively,
;;;	- or storing a copy of the list in the displayer's help-key slot?

(defun help (pane type &optional (help-key (capi:help-key pane)))
 ;;; Force displaying help when this is not done by CAPI automatically.
  ;; Args: type Must not be :tooltip
  (let ((interface (capi:top-level-interface pane)))
    (when-let (capi:interface-help-callback (capi:interface-help-callback interface))
      (funcall capi:interface-help-callback interface pane type help-key))))

(defun help-key-tooltip (help-key)
 ;;; Arg: help-key  Should not be :parent - we must have moved outward by means of
  ;;                (capi:help-key (capi:element-parent)) or cascading help-callback.
  ;; Value: String or NIL.
  (cond ((eq help-key :input) #1=#L"Input some value")
        ((stringp help-key) help-key)
        ((consp help-key) (let* ((rest (cdr help-key))
                                 (tooltip (if (consp rest)
                                              (cdr rest) ; (topic-id context-id . tooltip)
                                              rest)))	 ; (topic-id . tooltip)
                            (when (stringp tooltip)	 ; for sure
                              tooltip)))))

(defgeneric help-callback (interface pane type &optional help-key)
 ;;; Args: type	     For :tooltip, methods are allowed only to return something,
  ;;		     but not to display anything.
  ;;	   help-key  The value of capi:help-key of the pane if it is a string
  ;;                 or capi:capi-object-name of the pane if capi:help-key is T.
  ;; Value: See CAPI documentation.
 (:method (interface pane type &optional help-key)
  (declare (ignore interface pane type help-key)))

 ;; Tooltip for an entire interface (even non-top-level) is an absurd by default
 (:method (interface pane (type (eql :tooltip)) &optional help-key)
  (declare (ignore interface pane type))
  (if *show-tooltips*
      (if (eq (first-or-self help-key) :parent)
          (help-callback interface (capi:element-parent pane) type)
          (help-key-tooltip help-key))
      nil))
 (:method (interface (pane capi:interface) (type (eql :tooltip)) &optional help-key)
  (declare (ignore interface pane type help-key))
  nil)

 (:method (interface pane (type (eql :pointer-documentation-enter)) &optional help-key)
  ;; Auto called only for menus with item's help-key or name passed as help-key.
  ;; Q: Exclude rude auto-generated keys that come with (e.g.) the OK button?
   ;(and help-key (not (equal help-key (capi:capi-object-name pane))))
  (declare (ignore type #-debug pane))
  (when help-key
    (when-let (tooltip (or (help-key-tooltip help-key)
                           #+debug (format nil "** Help for ~s, help-key  ~s"
                                           pane help-key)))
      (message-area-format interface tooltip))))

 (:method (interface pane (type (eql :pointer-documentation-leave)) &optional help-key)
  (declare (ignore pane type))
  (when help-key
    (message-area-format interface +null-string+))))

(let ((lw:*handle-warn-on-redefinition* :quiet))
  (defmethod CAPI:HELP-KEY ((object capi:capi-object))
    nil))

(defmethod help-callback (interface pane (type (eql :help))
                          &optional (help-key (if pane (capi:help-key pane))))
 ;;; CAUTION: Pressing F1 does not invoke this callback when help-key is NIL!
  ;; CAPI NB:
  ;; 1) Is not invoked in any way for the following pane classes:
  ;;	- output-pane, seems cannot get focus.
  ;; 2) Is not invoked by means of F1 for the following pane classes:
  ;;	- extended-selection-tree-view.
  (debug-format :help "Help on ~s, key ~s" pane help-key)
  (cond ((null help-key)				; for compatibility with F1,
         nil)						; we must ignore such request
        ((eq (first-or-self help-key) :parent)
         (help-callback interface (capi:element-parent pane) type))
        ;((setq type (help-key-tooltip help-key))
        ; (let ((*show-tooltips* t))
        ;   (capi:display-tooltip pane :text type)	; works only for output-pane
        ;   (CAPI::PERFORM-CAPI-HELP-CALLBACK interface pane :tooltip t)))
        (t
         (beep-box #L"~@[~a~2%~]Context help is not implemented for ~s."
                   :format-arguments (list (help-key-tooltip help-key)
                                           (or (object-title (or pane interface))
                                               (and pane (capi:capi-object-name pane))
                                               help-key))
                   :icon :info  :owner interface))))
